import unittest
import operator
import collections

class NotImplementedByController(unittest.SkipTest, NotImplementedError):
    def __str__(self):
        return 'Not implemented by controller: {}'.format(self.args[0])

class ImplementationChoice(unittest.SkipTest):
    def __str__(self):
        return 'Choice in the implementation makes it impossible to ' \
                'perform a test: {}'.format(self.args[0])

class OptionalExtensionNotSupported(unittest.SkipTest):
    def __str__(self):
        return 'Unsupported extension: {}'.format(self.args[0])

class OptionalSaslMechanismNotSupported(unittest.SkipTest):
    def __str__(self):
        return 'Unsupported SASL mechanism: {}'.format(self.args[0])

class NotRequiredBySpecifications(unittest.SkipTest):
    def __str__(self):
        return 'Tests not required by the set of tested specification(s).'

class SkipStrictTest(unittest.SkipTest):
    def __str__(self):
        return 'Tests not required because strict tests are disabled.'

class TextTestResult(unittest.TextTestResult):
    def getDescription(self, test):
        if hasattr(test, 'description'):
            doc_first_lines = test.description()
        else:
            doc_first_lines = test.shortDescription()
        return '\n'.join((str(test), doc_first_lines or ''))

class TextTestRunner(unittest.TextTestRunner):
    """Small wrapper around unittest.TextTestRunner that reports the
    number of tests that were skipped because the software does not support
    an optional feature."""
    resultclass = TextTestResult

    def run(self, test):
        result = super().run(test)
        assert self.resultclass is TextTestResult
        if result.skipped:
            print()
            print('Some tests were skipped because the following optional '
                    'specifications/mechanisms are not supported:')
            msg_to_count = collections.defaultdict(lambda: 0)
            for (test, msg) in result.skipped:
                msg_to_count[msg] += 1
            for (msg, count) in sorted(msg_to_count.items()):
                print('\t{} ({} test(s))'.format(msg, count))
        return result



import os
import shutil
import socket
import tempfile
import time
import subprocess
import psutil

from .runner import NotImplementedByController

class _BaseController:
    """Base class for software controllers.

    A software controller is an object that handles configuring and running
    a process (eg. a server or a client), as well as sending it instructions
    that are not part of the IRC specification."""
    pass

class DirectoryBasedController(_BaseController):
    """Helper for controllers whose software configuration is based on an
    arbitrary directory."""
    def __init__(self):
        super().__init__()
        self.directory = None
        self.proc = None

    def kill_proc(self):
        """Terminates the controlled process, waits for it to exit, and
        eventually kills it."""
        self.proc.terminate()
        try:
            self.proc.wait(5)
        except subprocess.TimeoutExpired:
            self.proc.kill()
        self.proc = None
    def kill(self):
        """Calls `kill_proc` and cleans the configuration."""
        if self.proc:
            self.kill_proc()
        if self.directory:
            shutil.rmtree(self.directory)
    def open_file(self, name, mode='a'):
        """Open a file in the configuration directory."""
        assert self.directory
        if os.sep in name:
            dir_ = os.path.join(self.directory, os.path.dirname(name))
            if not os.path.isdir(dir_):
                os.makedirs(dir_)
            assert os.path.isdir(dir_)
        return open(os.path.join(self.directory, name), mode)
    def create_config(self):
        self.directory = tempfile.mkdtemp()

    def gen_ssl(self):
        self.csr_path = os.path.join(self.directory, 'ssl.csr')
        self.key_path = os.path.join(self.directory, 'ssl.key')
        self.pem_path = os.path.join(self.directory, 'ssl.pem')
        self.dh_path = os.path.join(self.directory, 'dh.pem')
        subprocess.check_output(['openssl', 'req', '-new', '-newkey', 'rsa',
            '-nodes', '-out', self.csr_path, '-keyout', self.key_path,
            '-batch'],
            stderr=subprocess.DEVNULL)
        subprocess.check_output(['openssl', 'x509', '-req',
            '-in', self.csr_path, '-signkey', self.key_path,
            '-out', self.pem_path],
            stderr=subprocess.DEVNULL)
        subprocess.check_output(['openssl', 'dhparam',
            '-out', self.dh_path, '128'],
            stderr=subprocess.DEVNULL)

class BaseClientController(_BaseController):
    """Base controller for IRC clients."""
    def run(self, hostname, port, auth):
        raise NotImplementedError()

class BaseServerController(_BaseController):
    """Base controller for IRC server."""
    port_open = False
    def run(self, hostname, port, password,
            valid_metadata_keys, invalid_metadata_keys):
        raise NotImplementedError()
    def registerUser(self, case, username, password=None):
        raise NotImplementedByController('registration')
    def wait_for_port(self):
        while not self.port_open:
            time.sleep(0.1)
            for conn in psutil.Process(self.proc.pid).connections():
                if conn.laddr[1] == self.port:
                    self.port_open = True




import enum

@enum.unique
class Specifications(enum.Enum):
    RFC1459 = 'RFC1459'
    RFC2812 = 'RFC2812'
    IRC301 = 'IRCv3.1'
    IRC302 = 'IRCv3.2'

    @classmethod
    def of_name(cls, name):
        name = name.upper()
        for spec in cls:
            if spec.value.upper() == name:
                return spec
        raise ValueError(name)



import os
import time
import shutil
import tempfile
import subprocess

from irctest import client_mock
from irctest import authentication
from irctest.basecontrollers import NotImplementedByController
from irctest.basecontrollers import BaseServerController, DirectoryBasedController

TEMPLATE_CONFIG = """
serverinfo {{
    name = "My.Little.Server";
    sid = "42X";
    description = "test server";
{ssl_config}
}};
listen {{
    defer_accept = yes;

    host = "{hostname}";
    port = {port};
}};
auth {{
    user = "*";
    flags = exceed_limit;
    {password_field}
}};
channel {{
    disable_local_channels = no;
    no_create_on_split = no;
    no_join_on_split = no;
    displayed_usercount = 0;
}};
"""

TEMPLATE_SSL_CONFIG = """
    ssl_private_key = "{key_path}";
    ssl_cert = "{pem_path}";
    ssl_dh_params = "{dh_path}";
"""


class CharybdisController(BaseServerController, DirectoryBasedController):
    software_name = 'Charybdis'
    supported_sasl_mechanisms = set()
    def create_config(self):
        super().create_config()
        with self.open_file('server.conf'):
            pass

    def run(self, hostname, port, password=None, ssl=False,
            valid_metadata_keys=None, invalid_metadata_keys=None):
        if valid_metadata_keys or invalid_metadata_keys:
            raise NotImplementedByController(
                    'Defining valid and invalid METADATA keys.')
        assert self.proc is None
        self.create_config()
        self.port = port
        password_field = 'password = "{}";'.format(password) if password else ''
        if ssl:
            self.gen_ssl()
            ssl_config = TEMPLATE_SSL_CONFIG.format(
                    key_path=self.key_path,
                    pem_path=self.pem_path,
                    dh_path=self.dh_path,
                    )
        else:
            ssl_config = ''
        with self.open_file('server.conf') as fd:
            fd.write(TEMPLATE_CONFIG.format(
                hostname=hostname,
                port=port,
                password_field=password_field,
                ssl_config=ssl_config,
                ))
        self.proc = subprocess.Popen(['ircd', '-foreground',
            '-configfile', os.path.join(self.directory, 'server.conf'),
            '-pidfile', os.path.join(self.directory, 'server.pid'),
            ],
            stderr=subprocess.DEVNULL
            )


def get_irctest_controller_class():
    return CharybdisController



import os
import time
import shutil
import tempfile
import subprocess

from irctest import authentication
from irctest.basecontrollers import NotImplementedByController
from irctest.basecontrollers import BaseServerController, DirectoryBasedController

TEMPLATE_CONFIG = """
<bind address="{hostname}" port="{port}" type="clients">
{ssl_config}
<module name="cap">
<module name="ircv3">
<module name="ircv3_capnotify">
<module name="ircv3_echomessage">
<module name="namesx"> # For multi-prefix
<connect allow="*"
    resolvehostnames="no" # Faster
    {password_field}>
<log method="file" type="*" level="debug" target="/tmp/ircd-{port}.log">
"""

TEMPLATE_SSL_CONFIG = """
<module name="ssl_openssl">
<openssl certfile="{pem_path}" keyfile="{key_path}" dhfile="{dh_path}" hash="sha1">
"""

class InspircdController(BaseServerController, DirectoryBasedController):
    software_name = 'InspIRCd'
    supported_sasl_mechanisms = set()
    def create_config(self):
        super().create_config()
        with self.open_file('server.conf'):
            pass

    def run(self, hostname, port, password=None, ssl=False,
            restricted_metadata_keys=None,
            valid_metadata_keys=None, invalid_metadata_keys=None):
        if valid_metadata_keys or invalid_metadata_keys:
            raise NotImplementedByController(
                    'Defining valid and invalid METADATA keys.')
        assert self.proc is None
        self.port = port
        self.create_config()
        password_field = 'password="{}"'.format(password) if password else ''
        if ssl:
            self.gen_ssl()
            ssl_config = TEMPLATE_SSL_CONFIG.format(
                    key_path=self.key_path,
                    pem_path=self.pem_path,
                    dh_path=self.dh_path,
                    )
        else:
            ssl_config = ''
        with self.open_file('server.conf') as fd:
            fd.write(TEMPLATE_CONFIG.format(
                hostname=hostname,
                port=port,
                password_field=password_field,
                ssl_config=ssl_config
                ))
        self.proc = subprocess.Popen(['inspircd', '--nofork', '--config',
            os.path.join(self.directory, 'server.conf')],
            stdout=subprocess.DEVNULL
            )

def get_irctest_controller_class():
    return InspircdController




"""
<http://ircv3.net/specs/extensions/account-tag-3.2.html>
"""

from irctest import cases
from irctest.client_mock import NoMessageException
from irctest.basecontrollers import NotImplementedByController

class AccountTagTestCase(cases.BaseServerTestCase, cases.OptionalityHelper):
    def connectRegisteredClient(self, nick):
        self.addClient()
        self.sendLine(2, 'CAP LS 302')
        capabilities = self.getCapLs(2)
        assert 'sasl' in capabilities
        self.sendLine(2, 'AUTHENTICATE PLAIN')
        m = self.getMessage(2, filter_pred=lambda m:m.command != 'NOTICE')
        self.assertMessageEqual(m, command='AUTHENTICATE', params=['+'],
                fail_msg='Sent “AUTHENTICATE PLAIN”, server should have '
                'replied with “AUTHENTICATE +”, but instead sent: {msg}')
        self.sendLine(2, 'AUTHENTICATE amlsbGVzAGppbGxlcwBzZXNhbWU=')
        m = self.getMessage(2, filter_pred=lambda m:m.command != 'NOTICE')
        self.assertMessageEqual(m, command='900',
                fail_msg='Did not send 900 after correct SASL authentication.')
        self.sendLine(2, 'USER f * * :Realname')
        self.sendLine(2, 'NICK {}'.format(nick))
        self.sendLine(2, 'CAP END')
        self.skipToWelcome(2)

    @cases.SpecificationSelector.requiredBySpecification('IRCv3.2')
    @cases.OptionalityHelper.skipUnlessHasMechanism('PLAIN')
    def testPrivmsg(self):
        self.connectClient('foo', capabilities=['account-tag'],
                skip_if_cap_nak=True)
        self.getMessages(1)
        self.controller.registerUser(self, 'jilles', 'sesame')
        self.connectRegisteredClient('bar')
        self.sendLine(2, 'PRIVMSG foo :hi')
        self.getMessages(2)
        m = self.getMessage(1)
        self.assertMessageEqual(m, command='PRIVMSG', # RPL_MONONLINE
                fail_msg='Sent non-730 (RPL_MONONLINE) message after '
                '“bar” sent a PRIVMSG: {msg}')
        self.assertIn('account', m.tags, m,
                fail_msg='PRIVMSG by logged in nick '
                'does not contain an account tag: {msg}')
        self.assertEqual(m.tags['account'], 'jilles', m,
                fail_msg='PRIVMSG by logged in nick '
                'does not contain the correct account tag (should be '
                '“jilles”): {msg}')


    @cases.SpecificationSelector.requiredBySpecification('IRCv3.2')
    @cases.OptionalityHelper.skipUnlessHasMechanism('PLAIN')
    def testMonitor(self):
        self.connectClient('foo', capabilities=['account-tag'],
                skip_if_cap_nak=True)
        if 'MONITOR' not in self.server_support:
            raise NotImplementedByController('MONITOR')
        self.sendLine(1, 'MONITOR + bar')
        self.getMessages(1)
        self.controller.registerUser(self, 'jilles', 'sesame')
        self.connectRegisteredClient('bar')
        m = self.getMessage(1)
        self.assertMessageEqual(m, command='730', # RPL_MONONLINE
                fail_msg='Sent non-730 (RPL_MONONLINE) message after '
                'monitored nick “bar” connected: {msg}')
        self.assertEqual(len(m.params), 2, m,
                fail_msg='Invalid number of params of RPL_MONONLINE: {msg}')
        self.assertEqual(m.params[1].split('!')[0], 'bar',
                fail_msg='730 (RPL_MONONLINE) with bad target after “bar” '
                'connects: {msg}')
        self.assertIn('account', m.tags, m,
                fail_msg='730 (RPL_MONONLINE) sent because of logged in nick '
                'does not contain an account tag: {msg}')
        self.assertEqual(m.tags['account'], 'jilles', m,
                fail_msg='730 (RPL_MONONLINE) sent because of logged in nick '
                'does not contain the correct account tag (should be '
                '“jilles”): {msg}')



import os
import time
import subprocess

from irctest.basecontrollers import NotImplementedByController
from irctest.basecontrollers import BaseServerController, DirectoryBasedController

TEMPLATE_CONFIG = """
clients:
  # ping_frequency - client ping frequency
  ping_frequency:
    minutes: 1000

  # ping_timeout - ping timeout length
  ping_timeout:
    seconds: 10
data:
  format: json
  filename: {directory}/data.json
  save_frequency:
    minutes: 5
extensions:
- mammon.ext.rfc1459.42
- mammon.ext.rfc1459.ident
- mammon.ext.ircv3.account_notify
- mammon.ext.ircv3.server_time
- mammon.ext.ircv3.echo_message
- mammon.ext.ircv3.register
- mammon.ext.ircv3.sasl
- mammon.ext.misc.nopost
metadata:
  restricted_keys:
{restricted_keys}
  whitelist:
{authorized_keys}
monitor:
  limit: 20
motd:
  - "Hi"
limits:
  foo: bar
listeners:
- {{"host": "{hostname}", "port": {port}, "ssl": false}}
logs:
  {{
  }}
register:
  enabled_callbacks:
  - none
  verify_timeout:
    days: 1
roles:
  "placeholder":
    title: "Just a placeholder"
server:
  name: MyLittleServer
  network: MyLittleNetwork
  recvq_len: 20
"""

def make_list(l):
    return '\n'.join(map('  - {}'.format, l))

class MammonController(BaseServerController, DirectoryBasedController):
    software_name = 'Mammon'
    supported_sasl_mechanisms = {
            'PLAIN', 'ECDSA-NIST256P-CHALLENGE',
            }
    def create_config(self):
        super().create_config()
        with self.open_file('server.conf'):
            pass

    def kill_proc(self):
        # Mammon does not seem to handle SIGTERM very well
        self.proc.kill()

    def run(self, hostname, port, password=None, ssl=False,
            restricted_metadata_keys=(),
            valid_metadata_keys=(), invalid_metadata_keys=()):
        if password is not None:
            raise NotImplementedByController('PASS command')
        if ssl:
            raise NotImplementedByController('SSL')
        assert self.proc is None
        self.port = port
        self.create_config()
        with self.open_file('server.yml') as fd:
            fd.write(TEMPLATE_CONFIG.format(
                directory=self.directory,
                hostname=hostname,
                port=port,
                authorized_keys=make_list(valid_metadata_keys),
                restricted_keys=make_list(restricted_metadata_keys),
                ))
        #with self.open_file('server.yml', 'r') as fd:
        #    print(fd.read())
        self.proc = subprocess.Popen(['mammond', '--nofork', #'--debug',
            '--config', os.path.join(self.directory, 'server.yml')])

    def registerUser(self, case, username, password=None):
        # XXX: Move this somewhere else when
        # https://github.com/ircv3/ircv3-specifications/pull/152 becomes
        # part of the specification
        client = case.addClient(show_io=False)
        case.sendLine(client, 'CAP LS 302')
        case.sendLine(client, 'NICK registration_user')
        case.sendLine(client, 'USER r e g :user')
        case.sendLine(client, 'CAP END')
        while case.getRegistrationMessage(client).command != '001':
            pass
        list(case.getMessages(client))
        case.sendLine(client, 'REG CREATE {} passphrase {}'.format(
            username, password))
        msg = case.getMessage(client)
        assert msg.command == '920', msg
        list(case.getMessages(client))
        case.removeClient(client)

def get_irctest_controller_class():
    return MammonController



"""
Tests section 4.1 of RFC 1459.
<https://tools.ietf.org/html/rfc1459#section-4.1>
"""

from irctest import cases
from irctest import authentication
from irctest.irc_utils.message_parser import Message
from irctest.basecontrollers import NotImplementedByController
from irctest.client_mock import ConnectionClosed

class PasswordedConnectionRegistrationTestCase(cases.BaseServerTestCase):
    password = 'testpassword'
    @cases.SpecificationSelector.requiredBySpecification('RFC1459', 'RFC2812')
    def testPassBeforeNickuser(self):
        self.addClient()
        self.sendLine(1, 'PASS {}'.format(self.password))
        self.sendLine(1, 'NICK foo')
        self.sendLine(1, 'USER username * * :Realname')

        m = self.getRegistrationMessage(1)
        self.assertMessageEqual(m, command='001',
                fail_msg='Did not get 001 after correct PASS+NICK+USER: {msg}')

    @cases.SpecificationSelector.requiredBySpecification('RFC1459', 'RFC2812')
    def testNoPassword(self):
        self.addClient()
        self.sendLine(1, 'NICK foo')
        self.sendLine(1, 'USER username * * :Realname')
        m = self.getRegistrationMessage(1)
        self.assertNotEqual(m.command, '001',
                msg='Got 001 after NICK+USER but missing PASS')

    @cases.SpecificationSelector.requiredBySpecification('RFC1459', 'RFC2812')
    def testPassAfterNickuser(self):
        """“The password can and must be set before any attempt to register
        the connection is made.”
        -- <https://tools.ietf.org/html/rfc1459#section-4.1.1>

        “The optional password can and MUST be set before any attempt to
        register the connection is made.
        Currently this requires that user send a PASS command before
        sending the NICK/USER combination.”
        -- <https://tools.ietf.org/html/rfc2812#section-3.1.1>
        """
        self.addClient()
        self.sendLine(1, 'NICK foo')
        self.sendLine(1, 'USER username * * :Realname')
        self.sendLine(1, 'PASS {}'.format(self.password))
        m = self.getRegistrationMessage(1)
        self.assertNotEqual(m.command, '001',
                'Got 001 after PASS sent after NICK+USER')

class ConnectionRegistrationTestCase(cases.BaseServerTestCase):
    @cases.SpecificationSelector.requiredBySpecification('RFC1459')
    def testQuitDisconnects(self):
        """“The server must close the connection to a client which sends a
        QUIT message.”
        -- <https://tools.ietf.org/html/rfc1459#section-4.1.3>
        """
        self.connectClient('foo')
        self.getMessages(1)
        self.sendLine(1, 'QUIT')
        with self.assertRaises(ConnectionClosed):
            self.getMessages(1) # Fetch remaining messages
            self.getMessages(1)

    @cases.SpecificationSelector.requiredBySpecification('RFC2812')
    def testQuitErrors(self):
        """“A client session is terminated with a quit message.  The server
        acknowledges this by sending an ERROR message to the client.”
        -- <https://tools.ietf.org/html/rfc2812#section-3.1.7>
        """
        self.connectClient('foo')
        self.getMessages(1)
        self.sendLine(1, 'QUIT')
        try:
            commands = {m.command for m in self.getMessages(1)}
        except ConnectionClosed:
            assert False, 'Connection closed without ERROR.'
        self.assertIn('ERROR', commands,
                fail_msg='Did not receive ERROR as a reply to QUIT.')


    def testNickCollision(self):
        """A user connects and requests the same nickname as an already
        registered user.
        """
        self.connectClient('foo')
        self.addClient()
        self.sendLine(2, 'NICK foo')
        self.sendLine(2, 'USER username * * :Realname')
        m = self.getRegistrationMessage(2)
        self.assertNotEqual(m.command, '001',
                'Received 001 after registering with the nick of a '
                'registered user.')

    def testEarlyNickCollision(self):
        """Two users register simultaneously with the same nick."""
        self.addClient()
        self.addClient()
        self.sendLine(1, 'NICK foo')
        self.sendLine(2, 'NICK foo')
        self.sendLine(1, 'USER username * * :Realname')
        self.sendLine(2, 'USER username * * :Realname')
        m1 = self.getRegistrationMessage(1)
        m2 = self.getRegistrationMessage(2)
        self.assertNotEqual((m1.command, m2.command), ('001', '001'),
                'Two concurrently registering requesting the same nickname '
                'both got 001.')

    @cases.SpecificationSelector.requiredBySpecification('IRCv3.1', 'IRCv3.2')
    def testIrc301CapLs(self):
        """IRCv3.1: “The LS subcommand is used to list the capabilities
        supported by the server. The client should send an LS subcommand with
        no other arguments to solicit a list of all capabilities.”
        -- <http://ircv3.net/specs/core/capability-negotiation-3.1.html#the-cap-ls-subcommand>

        IRCv3.2: “Servers MUST NOT send messages described by this document if
        the client only supports version 3.1.”
        -- <http://ircv3.net/specs/core/capability-negotiation-3.2.html#version-in-cap-ls>
        """
        self.addClient()
        self.sendLine(1, 'CAP LS')
        m = self.getRegistrationMessage(1)
        self.assertNotEqual(m.params[2], '*', m,
                fail_msg='Server replied with multi-line CAP LS to a '
                '“CAP LS” (ie. IRCv3.1) request: {msg}')
        self.assertFalse(any('=' in cap for cap in m.params[2].split()),
                'Server replied with a name-value capability in '
                'CAP LS reply as a response to “CAP LS” (ie. IRCv3.1) '
                'request: {}'.format(m))

    @cases.SpecificationSelector.requiredBySpecification('IRCv3.1')
    def testEmptyCapList(self):
        """“If no capabilities are active, an empty parameter must be sent.”
        -- <http://ircv3.net/specs/core/capability-negotiation-3.1.html#the-cap-list-subcommand>
        """
        self.addClient()
        self.sendLine(1, 'CAP LIST')
        m = self.getRegistrationMessage(1)
        self.assertMessageEqual(m, command='CAP', params=['*', 'LIST', ''],
                fail_msg='Sending “CAP LIST” as first message got a reply '
                'that is not “CAP * LIST :”: {msg}')



from irctest import cases
from irctest.irc_utils.message_parser import Message

class CapTestCase(cases.BaseClientTestCase, cases.ClientNegociationHelper):
    @cases.SpecificationSelector.requiredBySpecification('IRCv3.1', 'IRCv3.2')
    def testSendCap(self):
        """Send CAP LS 302 and read the result."""
        self.readCapLs()

    @cases.SpecificationSelector.requiredBySpecification('IRCv3.1', 'IRCv3.2')
    def testEmptyCapLs(self):
        """Empty result to CAP LS. Client should send CAP END."""
        m = self.negotiateCapabilities([])
        self.assertEqual(m, Message([], None, 'CAP', ['END']))



"""
<http://ircv3.net/specs/extensions/tls-3.1.html>
"""

from irctest import cases
from irctest.basecontrollers import NotImplementedByController

class StarttlsFailTestCase(cases.BaseServerTestCase):
    @cases.SpecificationSelector.requiredBySpecification('IRCv3.1')
    def testStarttlsRequestTlsFail(self):
        """<http://ircv3.net/specs/extensions/tls-3.1.html>
        """
        self.addClient()

        # TODO: check also without this
        self.sendLine(1, 'CAP LS')
        capabilities = self.getCapLs(1)
        if 'tls' not in capabilities:
            raise NotImplementedByController('tls')

        # TODO: check also without this
        self.sendLine(1, 'CAP REQ :tls')
        m = self.getRegistrationMessage(1)
        # TODO: Remove this once the trailing space issue is fixed in Charybdis
        # and Mammon:
        #self.assertMessageEqual(m, command='CAP', params=['*', 'ACK', 'tls'],
        #        fail_msg='Did not ACK capability `tls`: {msg}')
        self.sendLine(1, 'STARTTLS')
        m = self.getRegistrationMessage(1)
        self.assertMessageEqual(m, command='691',
                fail_msg='Did not respond to STARTTLS with 691 whereas '
                'SSL is not configured: {msg}.')

class StarttlsTestCase(cases.BaseServerTestCase):
    ssl = True
    def testStarttlsRequestTls(self):
        """<http://ircv3.net/specs/extensions/tls-3.1.html>
        """
        self.addClient()

        # TODO: check also without this
        self.sendLine(1, 'CAP LS')
        capabilities = self.getCapLs(1)
        if 'tls' not in capabilities:
            raise NotImplementedByController('tls')

        # TODO: check also without this
        self.sendLine(1, 'CAP REQ :tls')
        m = self.getRegistrationMessage(1)
        # TODO: Remove this one the trailing space issue is fixed in Charybdis
        # and Mammon:
        #self.assertMessageEqual(m, command='CAP', params=['*', 'ACK', 'tls'],
        #        fail_msg='Did not ACK capability `tls`: {msg}')
        self.sendLine(1, 'STARTTLS')
        m = self.getRegistrationMessage(1)
        self.assertMessageEqual(m, command='670',
                fail_msg='Did not respond to STARTTLS with 670: {msg}.')
        self.clients[1].starttls()
        self.sendLine(1, 'USER f * * :foo')
        self.sendLine(1, 'NICK foo')
        self.sendLine(1, 'CAP END')
        self.getMessages(1)



import ssl
import time
import socket
import tempfile
import unittest
import functools
import collections

import supybot.utils

from . import runner
from . import client_mock
from . import authentication
from .irc_utils import capabilities
from .irc_utils import message_parser
from .exceptions import ConnectionClosed
from .specifications import Specifications

class _IrcTestCase(unittest.TestCase):
    """Base class for test cases."""
    controllerClass = None # Will be set by __main__.py

    def description(self):
        method_doc = self._testMethodDoc
        if not method_doc:
            return ''
        return '\t'+supybot.utils.str.normalizeWhitespace(
                method_doc,
                removeNewline=False,
                ).strip().replace('\n ', '\n\t')

    def setUp(self):
        super().setUp()
        self.controller = self.controllerClass()
        self.inbuffer = []
        if self.show_io:
            print('---- new test ----')
    def assertMessageEqual(self, msg, subcommand=None, subparams=None,
            target=None, nick=None, fail_msg=None, extra_format=(),
            strip_first_param=False, **kwargs):
        """Helper for partially comparing a message.

        Takes the message as first arguments, and comparisons to be made
        as keyword arguments.

        Deals with subcommands (eg. `CAP`) if any of `subcommand`,
        `subparams`, and `target` are given."""
        fail_msg = fail_msg or '{msg}'
        for (key, value) in kwargs.items():
            if strip_first_param and key == 'params':
                value = value[1:]
            self.assertEqual(getattr(msg, key), value, msg, fail_msg,
                    extra_format=extra_format)
        if nick:
            self.assertNotEqual(msg.prefix, None, msg, fail_msg)
            self.assertEqual(msg.prefix.split('!')[0], nick, msg, fail_msg)
        if subcommand is not None or subparams is not None:
            self.assertGreater(len(msg.params), 2, fail_msg)
            msg_target = msg.params[0]
            msg_subcommand = msg.params[1]
            msg_subparams = msg.params[2:]
            if subcommand:
                self.assertEqual(msg_subcommand, subcommand, msg, fail_msg,
                        extra_format=extra_format)
            if subparams is not None:
                self.assertEqual(msg_subparams, subparams, msg, fail_msg,
                        extra_format=extra_format)

    def assertIn(self, item, list_, msg=None, fail_msg=None, extra_format=()):
        if fail_msg:
            fail_msg = fail_msg.format(*extra_format,
                    item=item, list=list_, msg=msg)
        super().assertIn(item, list_, fail_msg)
    def assertNotIn(self, item, list_, msg=None, fail_msg=None, extra_format=()):
        if fail_msg:
            fail_msg = fail_msg.format(*extra_format,
                    item=item, list=list_, msg=msg)
        super().assertNotIn(item, list_, fail_msg)
    def assertEqual(self, got, expects, msg=None, fail_msg=None, extra_format=()):
        if fail_msg:
            fail_msg = fail_msg.format(*extra_format,
                    got=got, expects=expects, msg=msg)
        super().assertEqual(got, expects, fail_msg)
    def assertNotEqual(self, got, expects, msg=None, fail_msg=None, extra_format=()):
        if fail_msg:
            fail_msg = fail_msg.format(*extra_format,
                    got=got, expects=expects, msg=msg)
        super().assertNotEqual(got, expects, fail_msg)

class BaseClientTestCase(_IrcTestCase):
    """Basic class for client tests. Handles spawning a client and exchanging
    messages with it."""
    nick = None
    user = None
    def setUp(self):
        super().setUp()
        self.conn = None
        self._setUpServer()
    def tearDown(self):
        if self.conn:
            self.conn.sendall(b'QUIT :end of test.')
        self.controller.kill()
        if self.conn:
            self.conn_file.close()
            self.conn.close()
        self.server.close()

    def _setUpServer(self):
        """Creates the server and make it listen."""
        self.server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.server.bind(('', 0)) # Bind any free port
        self.server.listen(1)
    def acceptClient(self, tls_cert=None, tls_key=None):
        """Make the server accept a client connection. Blocking."""
        (self.conn, addr) = self.server.accept()
        if tls_cert is None and tls_key is None:
            pass
        else:
            assert tls_cert and tls_key, \
                'tls_cert must be provided if and only if tls_key is.'
            with tempfile.NamedTemporaryFile('at') as certfile, \
                    tempfile.NamedTemporaryFile('at') as keyfile:
                certfile.write(tls_cert)
                certfile.seek(0)
                keyfile.write(tls_key)
                keyfile.seek(0)
                context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
                context.load_cert_chain(certfile=certfile.name, keyfile=keyfile.name)
                self.conn = context.wrap_socket(self.conn, server_side=True)
        self.conn_file = self.conn.makefile(newline='\r\n',
                encoding='utf8')

    def getLine(self):
        line = self.conn_file.readline()
        if self.show_io:
            print('{:.3f} C: {}'.format(time.time(), line.strip()))
        return line
    def getMessages(self, *args):
        lines = self.getLines(*args)
        return map(message_parser.parse_message, lines)
    def getMessage(self, *args, filter_pred=None):
        """Gets a message and returns it. If a filter predicate is given,
        fetches messages until the predicate returns a False on a message,
        and returns this message."""
        while True:
            line = self.getLine(*args)
            if not line:
                raise ConnectionClosed()
            msg = message_parser.parse_message(line)
            if not filter_pred or filter_pred(msg):
                return msg
    def sendLine(self, line):
        ret = self.conn.sendall(line.encode())
        assert ret is None
        if not line.endswith('\r\n'):
            ret = self.conn.sendall(b'\r\n')
            assert ret is None
        if self.show_io:
            print('{:.3f} S: {}'.format(time.time(), line.strip()))

class ClientNegociationHelper:
    """Helper class for tests handling capabilities negociation."""
    def readCapLs(self, auth=None, tls_config=None):
        (hostname, port) = self.server.getsockname()
        self.controller.run(
                hostname=hostname,
                port=port,
                auth=auth,
                tls_config=tls_config,
                )
        self.acceptClient()
        m = self.getMessage()
        self.assertEqual(m.command, 'CAP',
                'First message is not CAP LS.')
        if m.params == ['LS']:
            self.protocol_version = 301
        elif m.params == ['LS', '302']:
            self.protocol_version = 302
        elif m.params == ['END']:
            self.protocol_version = None
        else:
            raise AssertionError('Unknown CAP params: {}'
                    .format(m.params))

    def userNickPredicate(self, msg):
        """Predicate to be used with getMessage to handle NICK/USER
        transparently."""
        if msg.command == 'NICK':
            self.assertEqual(len(msg.params), 1, msg)
            self.nick = msg.params[0]
            return False
        elif msg.command == 'USER':
            self.assertEqual(len(msg.params), 4, msg)
            self.user = msg.params
            return False
        else:
            return True

    def negotiateCapabilities(self, caps, cap_ls=True, auth=None):
        """Performes a complete capability negociation process, without
        ending it, so the caller can continue the negociation."""
        if cap_ls:
            self.readCapLs(auth)
            if not self.protocol_version:
                # No negotiation.
                return
            self.sendLine('CAP * LS :{}'.format(' '.join(caps)))
        capability_names = frozenset(capabilities.cap_list_to_dict(caps))
        self.acked_capabilities = set()
        while True:
            m = self.getMessage(filter_pred=self.userNickPredicate)
            if m.command != 'CAP':
                return m
            self.assertGreater(len(m.params), 0, m)
            if m.params[0] == 'REQ':
                self.assertEqual(len(m.params), 2, m)
                requested = frozenset(m.params[1].split())
                if not requested.issubset(capability_names):
                    self.sendLine('CAP {} NAK :{}'.format(
                        self.nick or '*',
                        m.params[1][0:100]))
                else:
                    self.sendLine('CAP {} ACK :{}'.format(
                        self.nick or '*',
                        m.params[1]))
                    self.acked_capabilities.update(requested)
            else:
                return m


class BaseServerTestCase(_IrcTestCase):
    """Basic class for server tests. Handles spawning a server and exchanging
    messages with it."""
    password = None
    ssl = False
    valid_metadata_keys = frozenset()
    invalid_metadata_keys = frozenset()
    def setUp(self):
        super().setUp()
        self.server_support = {}
        self.find_hostname_and_port()
        self.controller.run(self.hostname, self.port, password=self.password,
                valid_metadata_keys=self.valid_metadata_keys,
                invalid_metadata_keys=self.invalid_metadata_keys,
                ssl=self.ssl)
        self.clients = {}
    def tearDown(self):
        self.controller.kill()
        for client in list(self.clients):
            self.removeClient(client)
    def find_hostname_and_port(self):
        """Find available hostname/port to listen on."""
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.bind(("",0))
        (self.hostname, self.port) = s.getsockname()
        s.close()

    def addClient(self, name=None, show_io=None):
        """Connects a client to the server and adds it to the dict.
        If 'name' is not given, uses the lowest unused non-negative integer."""
        self.controller.wait_for_port()
        if not name:
            name = max(map(int, list(self.clients)+[0]))+1
        show_io = show_io if show_io is not None else self.show_io
        self.clients[name] = client_mock.ClientMock(name=name,
                show_io=show_io)
        self.clients[name].connect(self.hostname, self.port)
        return name


    def removeClient(self, name):
        """Disconnects the client, without QUIT."""
        assert name in self.clients
        self.clients[name].disconnect()
        del self.clients[name]

    def getMessages(self, client, **kwargs):
        return self.clients[client].getMessages(**kwargs)
    def getMessage(self, client, **kwargs):
        return self.clients[client].getMessage(**kwargs)
    def getRegistrationMessage(self, client):
        """Filter notices, do not send pings."""
        return self.getMessage(client, synchronize=False,
                filter_pred=lambda m:m.command != 'NOTICE')
    def sendLine(self, client, line):
        return self.clients[client].sendLine(line)

    def getCapLs(self, client, as_list=False):
        """Waits for a CAP LS block, parses all CAP LS messages, and return
        the dict capabilities, with their values.

        If as_list is given, returns the raw list (ie. key/value not split)
        in case the order matters (but it shouldn't)."""
        caps = []
        while True:
            m = self.getRegistrationMessage(client)
            self.assertMessageEqual(m, command='CAP', subcommand='LS')
            if m.params[2] == '*':
                caps.extend(m.params[3].split())
            else:
                caps.extend(m.params[2].split())
                if not as_list:
                    caps = capabilities.cap_list_to_dict(caps)
                return caps

    def assertDisconnected(self, client):
        try:
            self.getLines(client)
            self.sendLine(client, 'PING foo')
            while True:
                l = self.getLine(client)
                self.assertNotEqual(line, '')
                m = message_parser.parse_message(l)
                self.assertNotEqual(m.command, 'PONG',
                        'Client not disconnected.')
        except socket.error:
            del self.clients[client]
            return
        else:
            raise AssertionError('Client not disconnected.')


    def skipToWelcome(self, client):
        """Skip to the point where we are registered
        <https://tools.ietf.org/html/rfc2812#section-3.1>
        """
        while True:
            m = self.getMessage(client, synchronize=False)
            if m.command == '001':
                return m
    def connectClient(self, nick, name=None, capabilities=None,
            skip_if_cap_nak=False):
        client = self.addClient(name)
        if capabilities is not None:
            self.sendLine(client, 'CAP REQ :{}'.format(' '.join(capabilities)))
            m = self.getRegistrationMessage(client)
            try:
                self.assertMessageEqual(m, command='CAP',
                        fail_msg='Expected CAP ACK, got: {msg}')
                self.assertEqual(m.params[1], 'ACK', m,
                        fail_msg='Expected CAP ACK, got: {msg}')
            except AssertionError:
                if skip_if_cap_nak:
                    raise runner.NotImplementedByController(
                            ', '.join(capabilities))
                else:
                    raise
            self.sendLine(client, 'CAP END')
        self.sendLine(client, 'NICK {}'.format(nick))
        self.sendLine(client, 'USER username * * :Realname')

        self.skipToWelcome(client)
        self.sendLine(client, 'PING foo')

        # Skip all that happy welcoming stuff
        while True:
            m = self.getMessage(client)
            if m.command == 'PONG':
                break
            elif m.command == '005':
                for param in m.params[1:-1]:
                    if '=' in param:
                        (key, value) = param.split('=')
                    else:
                        (key, value) = (param, None)
                    self.server_support[key] = value

    def joinClient(self, client, channel):
        self.sendLine(client, 'JOIN {}'.format(channel))
        received = {m.command for m in self.getMessages(client)}
        self.assertIn('366', received,
                fail_msg='Join to {} failed, {item} is not in the set of '
                'received responses: {list}',
                extra_format=(channel,))

class OptionalityHelper:
    def checkSaslSupport(self):
        if self.controller.supported_sasl_mechanisms:
            return
        raise runner.NotImplementedByController('SASL')

    def checkMechanismSupport(self, mechanism):
        if mechanism in self.controller.supported_sasl_mechanisms:
            return
        raise runner.OptionalSaslMechanismNotSupported(mechanism)

    def skipUnlessHasMechanism(mech):
        def decorator(f):
            @functools.wraps(f)
            def newf(self):
                self.checkMechanismSupport(mech)
                return f(self)
            return newf
        return decorator

    def skipUnlessHasSasl(f):
        @functools.wraps(f)
        def newf(self):
            self.checkSaslSupport()
            return f(self)
        return newf

class SpecificationSelector:

    def requiredBySpecification(*specifications, strict=False):
        specifications = frozenset(
                Specifications.of_name(s) if isinstance(s, str) else s
                for s in specifications)
        if None in specifications:
            raise ValueError('Invalid set of specifications: {}'
                    .format(specifications))
        def decorator(f):
            @functools.wraps(f)
            def newf(self):
                if specifications.isdisjoint(self.testedSpecifications):
                    raise runner.NotRequiredBySpecifications()
                if strict and not self.strictTests:
                    raise runner.SkipStrictTest()
                return f(self)
            return newf
        return decorator



import os
import subprocess

from irctest import authentication
from irctest.basecontrollers import NotImplementedByController
from irctest.basecontrollers import BaseClientController, DirectoryBasedController

TEMPLATE_CONFIG = """
supybot.directories.conf: {directory}/conf
supybot.directories.data: {directory}/data
supybot.directories.migrations: {directory}/migrations
supybot.log.level: DEBUG
supybot.log.stdout.level: {loglevel}

supybot.networks: testnet
supybot.networks.testnet.servers: {hostname}:{port}

supybot.protocols.ssl.verifyCertificates: True
supybot.networks.testnet.ssl: {enable_tls}
supybot.networks.testnet.ssl.serverFingerprints: {trusted_fingerprints}

supybot.networks.testnet.sasl.username: {username}
supybot.networks.testnet.sasl.password: {password}
supybot.networks.testnet.sasl.ecdsa_key: {directory}/ecdsa_key.pem
supybot.networks.testnet.sasl.mechanisms: {mechanisms}
"""

class LimnoriaController(BaseClientController, DirectoryBasedController):
    software_name = 'Limnoria'
    supported_sasl_mechanisms = {
            'PLAIN', 'ECDSA-NIST256P-CHALLENGE', 'SCRAM-SHA-256', 'EXTERNAL',
            }
    def create_config(self):
        super().create_config()
        with self.open_file('bot.conf'):
            pass
        with self.open_file('conf/users.conf'):
            pass

    def run(self, hostname, port, auth, tls_config):
        # Runs a client with the config given as arguments
        assert self.proc is None
        self.create_config()
        if auth:
            mechanisms = ' '.join(map(authentication.Mechanisms.as_string,
                auth.mechanisms))
            if auth.ecdsa_key:
                with self.open_file('ecdsa_key.pem') as fd:
                    fd.write(auth.ecdsa_key)
        else:
            mechanisms = ''
        with self.open_file('bot.conf') as fd:
            fd.write(TEMPLATE_CONFIG.format(
                directory=self.directory,
                loglevel='CRITICAL',
                hostname=hostname,
                port=port,
                username=auth.username if auth else '',
                password=auth.password if auth else '',
                mechanisms=mechanisms.lower(),
                enable_tls=tls_config.enable if tls_config else 'False',
                trusted_fingerprints=' '.join(tls_config.trusted_fingerprints) if tls_config else '',
                ))
        self.proc = subprocess.Popen(['supybot',
            os.path.join(self.directory, 'bot.conf')])

def get_irctest_controller_class():
    return LimnoriaController



"""
Tests METADATA features.
<http://ircv3.net/specs/core/metadata-3.2.html>
"""

from irctest import cases
from irctest.irc_utils.message_parser import Message

class MetadataTestCase(cases.BaseServerTestCase):
    valid_metadata_keys = {'valid_key1', 'valid_key2'}
    invalid_metadata_keys = {'invalid_key1', 'invalid_key2'}
    @cases.SpecificationSelector.requiredBySpecification('IRCv3.2')
    def testInIsupport(self):
        """“If METADATA is supported, it MUST be specified in RPL_ISUPPORT
        using the METADATA key.”
        -- <http://ircv3.net/specs/core/metadata-3.2.html>
        """
        self.addClient()
        self.sendLine(1, 'CAP LS 302')
        self.getCapLs(1)
        self.sendLine(1, 'USER foo foo foo :foo')
        self.sendLine(1, 'NICK foo')
        self.sendLine(1, 'CAP END')
        self.skipToWelcome(1)
        m = self.getMessage(1)
        while m.command != '005': # RPL_ISUPPORT
            m = self.getMessage(1)
        self.assertIn('METADATA', {x.split('=')[0] for x in m.params[1:-1]},
                fail_msg='{item} missing from RPL_ISUPPORT')
        self.getMessages(1)

    @cases.SpecificationSelector.requiredBySpecification('IRCv3.2')
    def testGetOneUnsetValid(self):
        """<http://ircv3.net/specs/core/metadata-3.2.html#metadata-get>
        """
        self.connectClient('foo')
        self.sendLine(1, 'METADATA * GET valid_key1')
        m = self.getMessage(1)
        self.assertMessageEqual(m, command='766', # ERR_NOMATCHINGKEY
                fail_msg='Did not reply with 766 (ERR_NOMATCHINGKEY) to a '
                'request to an unset valid METADATA key.')

    @cases.SpecificationSelector.requiredBySpecification('IRCv3.2')
    def testGetTwoUnsetValid(self):
        """“Multiple keys may be given. The response will be either RPL_KEYVALUE,
        ERR_KEYINVALID or ERR_NOMATCHINGKEY for every key in order.”
        -- <http://ircv3.net/specs/core/metadata-3.2.html#metadata-get>
        """
        self.connectClient('foo')
        self.sendLine(1, 'METADATA * GET valid_key1 valid_key2')
        m = self.getMessage(1)
        self.assertMessageEqual(m, command='766', # ERR_NOMATCHINGKEY
                fail_msg='Did not reply with 766 (ERR_NOMATCHINGKEY) to a '
                'request to two unset valid METADATA key: {msg}')
        self.assertEqual(m.params[1], 'valid_key1', m,
                fail_msg='Response to “METADATA * GET valid_key1 valid_key2” '
                'did not respond to valid_key1 first: {msg}')
        m = self.getMessage(1)
        self.assertMessageEqual(m, command='766', # ERR_NOMATCHINGKEY
                fail_msg='Did not reply with two 766 (ERR_NOMATCHINGKEY) to a '
                'request to two unset valid METADATA key: {msg}')
        self.assertEqual(m.params[1], 'valid_key2', m,
                fail_msg='Response to “METADATA * GET valid_key1 valid_key2” '
                'did not respond to valid_key2 as second response: {msg}')

    @cases.SpecificationSelector.requiredBySpecification('IRCv3.2')
    def testListNoSet(self):
        """“This subcommand MUST list all currently-set metadata keys along
        with their values. The response will be zero or more RPL_KEYVALUE
        events, following by RPL_METADATAEND event.”
        -- <http://ircv3.net/specs/core/metadata-3.2.html#metadata-list>
        """
        self.connectClient('foo')
        self.sendLine(1, 'METADATA * LIST')
        m = self.getMessage(1)
        self.assertMessageEqual(m, command='762', # RPL_METADATAEND
                fail_msg='Response to “METADATA * LIST” was not '
                '762 (RPL_METADATAEND) but: {msg}')

    @cases.SpecificationSelector.requiredBySpecification('IRCv3.2')
    def testListInvalidTarget(self):
        """“In case of invalid target RPL_METADATAEND MUST NOT be sent.”
        -- <http://ircv3.net/specs/core/metadata-3.2.html#metadata-list>
        """
        self.connectClient('foo')
        self.sendLine(1, 'METADATA foobar LIST')
        m = self.getMessage(1)
        self.assertMessageEqual(m, command='765', # ERR_TARGETINVALID
                fail_msg='Response to “METADATA <invalid target> LIST” was '
                'not 765 (ERR_TARGETINVALID) but: {msg}')
        commands = {m.command for m in self.getMessages(1)}
        self.assertNotIn('762', commands,
                fail_msg='Sent “METADATA <invalid target> LIST”, got 765 '
                '(ERR_TARGETINVALID), and then 762 (RPL_METADATAEND)')

    def assertSetValue(self, target, key, value, displayable_value=None):
        if displayable_value is None:
            displayable_value = value
        self.sendLine(1, 'METADATA {} SET {} :{}'.format(target, key, value))
        m = self.getMessage(1)
        self.assertMessageEqual(m, command='761', # RPL_KEYVALUE
                fail_msg='Did not reply with 761 (RPL_KEYVALUE) to a valid '
                '“METADATA * SET {} :{}”: {msg}',
                extra_format=(key, displayable_value,))
        self.assertEqual(m.params[1], 'valid_key1', m,
                fail_msg='Second param of 761 after setting “{expects}” to '
                '“{}” is not “{expects}”: {msg}.',
                extra_format=(displayable_value,))
        self.assertEqual(m.params[3], value, m,
                fail_msg='Fourth param of 761 after setting “{0}” to '
                '“{1}” is not “{1}”: {msg}.',
                extra_format=(key, displayable_value))
        m = self.getMessage(1)
        self.assertMessageEqual(m, command='762', # RPL_METADATAEND
                fail_msg='Did not send RPL_METADATAEND after setting '
                'a valid METADATA key.')
    def assertGetValue(self, target, key, value, displayable_value=None):
        self.sendLine(1, 'METADATA * GET {}'.format(key))
        m = self.getMessage(1)
        self.assertMessageEqual(m, command='761', # RPL_KEYVALUE
                fail_msg='Did not reply with 761 (RPL_KEYVALUE) to a valid '
                '“METADATA * GET” when the key is set is set: {msg}')
        self.assertEqual(m.params[1], key, m,
                fail_msg='Second param of 761 after getting “{expects}” '
                '(which is set) is not “{expects}”: {msg}.')
        self.assertEqual(m.params[3], value, m,
                fail_msg='Fourth param of 761 after getting “{0}” '
                '(which is set to “{1}”) is not ”{1}”: {msg}.',
                extra_format=(key, displayable_value))
    def assertSetGetValue(self, target, key, value, displayable_value=None):
        self.assertSetValue(target, key, value, displayable_value)
        self.assertGetValue(target, key, value, displayable_value)

    @cases.SpecificationSelector.requiredBySpecification('IRCv3.2')
    def testSetGetValid(self):
        """<http://ircv3.net/specs/core/metadata-3.2.html>
        """
        self.connectClient('foo')
        self.assertSetGetValue('*', 'valid_key1', 'myvalue')

    @cases.SpecificationSelector.requiredBySpecification('IRCv3.2')
    def testSetGetZeroCharInValue(self):
        """“Values are unrestricted, except that they MUST be UTF-8.”
        -- <http://ircv3.net/specs/core/metadata-3.2.html#metadata-restrictions>
        """
        self.connectClient('foo')
        self.assertSetGetValue('*', 'valid_key1', 'zero->\0<-zero',
                'zero->\\0<-zero')

    @cases.SpecificationSelector.requiredBySpecification('IRCv3.2')
    def testSetGetHeartInValue(self):
        """“Values are unrestricted, except that they MUST be UTF-8.”
        -- <http://ircv3.net/specs/core/metadata-3.2.html#metadata-restrictions>
        """
        heart = b'\xf0\x9f\x92\x9c'.decode()
        self.connectClient('foo')
        self.assertSetGetValue('*', 'valid_key1', '->{}<-'.format(heart),
                'zero->{}<-zero'.format(heart.encode()))

    @cases.SpecificationSelector.requiredBySpecification('IRCv3.2')
    def testSetInvalidUtf8(self):
        """“Values are unrestricted, except that they MUST be UTF-8.”
        -- <http://ircv3.net/specs/core/metadata-3.2.html#metadata-restrictions>
        """
        self.connectClient('foo')
        # Sending directly because it is not valid UTF-8 so Python would
        # not like it
        self.clients[1].conn.sendall(b'METADATA * SET valid_key1 '
                b':invalid UTF-8 ->\xc3<-\r\n')
        commands = {m.command for m in self.getMessages(1)}
        self.assertNotIn('761', commands, # RPL_KEYVALUE
                fail_msg='Setting METADATA key to a value containing invalid '
                'UTF-8 was answered with 761 (RPL_KEYVALUE)')
        self.clients[1].conn.sendall(b'METADATA * SET valid_key1 '
                b':invalid UTF-8: \xc3\r\n')
        commands = {m.command for m in self.getMessages(1)}
        self.assertNotIn('761', commands, # RPL_KEYVALUE
                fail_msg='Setting METADATA key to a value containing invalid '
                'UTF-8 was answered with 761 (RPL_KEYVALUE)')



"""
<http://ircv3.net/specs/extensions/echo-message-3.2.html>
"""

from irctest import cases
from irctest.basecontrollers import NotImplementedByController

class EchoMessageTestCase(cases.BaseServerTestCase):
    def _testEchoMessage(command, solo, server_time):
        @cases.SpecificationSelector.requiredBySpecification('IRCv3.2')
        def f(self):
            """<http://ircv3.net/specs/extensions/echo-message-3.2.html>
            """
            self.addClient()
            self.sendLine(1, 'CAP LS 302')
            capabilities = self.getCapLs(1)
            if 'echo-message' not in capabilities:
                raise NotImplementedByController('echo-message')
            if server_time and 'server-time' not in capabilities:
                raise NotImplementedByController('server-time')

            # TODO: check also without this
            self.sendLine(1, 'CAP REQ :echo-message{}'.format(
                ' server-time' if server_time else ''))
            m = self.getRegistrationMessage(1)
            # TODO: Remove this one the trailing space issue is fixed in Charybdis
            # and Mammon:
            #self.assertMessageEqual(m, command='CAP',
            #        params=['*', 'ACK', 'echo-message'] +
            #        (['server-time'] if server_time else []),
            #        fail_msg='Did not ACK advertised capabilities: {msg}')
            self.sendLine(1, 'USER f * * :foo')
            self.sendLine(1, 'NICK baz')
            self.sendLine(1, 'CAP END')
            self.skipToWelcome(1)
            self.getMessages(1)

            self.sendLine(1, 'JOIN #chan')

            if not solo:
                capabilities = ['server-time'] if server_time else []
                self.connectClient('qux', capabilities=capabilities)
                self.sendLine(2, 'JOIN #chan')

            # Synchronize and clean
            self.getMessages(1)
            if not solo:
                self.getMessages(2)
                self.getMessages(1)

            self.sendLine(1, '{} #chan :hello everyone'.format(command))
            m1 = self.getMessage(1)
            self.assertMessageEqual(m1, command=command,
                    params=['#chan', 'hello everyone'],
                    fail_msg='Did not echo “{} #chan :hello everyone”: {msg}',
                    extra_format=(command,))

            if not solo:
                m2 = self.getMessage(2)
                self.assertMessageEqual(m2, command=command,
                        params=['#chan', 'hello everyone'],
                        fail_msg='Did not propagate “{} #chan :hello everyone”: '
                        'after echoing it to the author: {msg}',
                        extra_format=(command,))
                self.assertEqual(m1.params, m2.params,
                        fail_msg='Parameters of forwarded and echoed '
                        'messages differ: {} {}',
                        extra_format=(m1, m2))
                if server_time:
                    self.assertEqual(m1.tags, m2.tags,
                            fail_msg='Tags of forwarded and echoed '
                            'messages differ: {} {}',
                            extra_format=(m1, m2))
        return f

    testEchoMessagePrivmsgNoServerTime = _testEchoMessage('PRIVMSG', False, False)
    testEchoMessagePrivmsgSolo = _testEchoMessage('PRIVMSG', True, True)
    testEchoMessagePrivmsg = _testEchoMessage('PRIVMSG', False, True)
    testEchoMessageNotice = _testEchoMessage('NOTICE', False, True)



import os
import tempfile
import subprocess

from irctest.basecontrollers import BaseClientController
from irctest.basecontrollers import NotImplementedByController

TEMPLATE_CONFIG = """
[core]
nick = Sopel
host = {hostname}
use_ssl = false
port = {port}
owner = me
channels = 
timeout = 5
auth_username = {username}
auth_password = {password}
{auth_method}
"""

class SopelController(BaseClientController):
    software_name = 'Sopel'
    supported_sasl_mechanisms = {
            'PLAIN',
            }
    def __init__(self):
        super().__init__()
        self.filename = next(tempfile._get_candidate_names()) + '.cfg'
        self.proc = None
    def kill(self):
        if self.proc:
            self.proc.kill()
        if self.filename:
            try:
                os.unlink(os.path.join(os.path.expanduser('~/.sopel/'),
                    self.filename))
            except OSError: # File does not exist
                pass

    def open_file(self, filename, mode='a'):
        return open(os.path.join(os.path.expanduser('~/.sopel/'), filename),
                mode)

    def create_config(self):
        with self.open_file(self.filename) as fd:
            pass

    def run(self, hostname, port, auth, tls_config):
        # Runs a client with the config given as arguments
        if tls_config is not None:
            raise NotImplementedByController(
                    'TLS configuration')
        assert self.proc is None
        self.create_config()
        with self.open_file(self.filename) as fd:
            fd.write(TEMPLATE_CONFIG.format(
                hostname=hostname,
                port=port,
                username=auth.username if auth else '',
                password=auth.password if auth else '',
                auth_method='auth_method = sasl' if auth else '',
                ))
        self.proc = subprocess.Popen(['sopel', '--quiet', '-c', self.filename])

def get_irctest_controller_class():
    return SopelController




"""
<http://ircv3.net/specs/core/monitor-3.2.html>
"""

from irctest import cases
from irctest.client_mock import NoMessageException
from irctest.basecontrollers import NotImplementedByController

class EchoMessageTestCase(cases.BaseServerTestCase):
    def check_server_support(self):
        if 'MONITOR' not in self.server_support:
            raise NotImplementedByController('MONITOR')

    def assertMononline(self, client, nick, m=None):
        if not m:
            m = self.getMessage(client)
        self.assertMessageEqual(m, command='730', # RPL_MONONLINE
                fail_msg='Sent non-730 (RPL_MONONLINE) message after '
                'monitored nick “{}” connected: {msg}',
                extra_format=(nick,))
        self.assertEqual(len(m.params), 2, m,
                fail_msg='Invalid number of params of RPL_MONONLINE: {msg}')
        self.assertEqual(m.params[1].split('!')[0], 'bar',
                fail_msg='730 (RPL_MONONLINE) with bad target after “{}” '
                'connects: {msg}',
                extra_format=(nick,))

    def assertMonoffline(self, client, nick, m=None):
        if not m:
            m = self.getMessage(client)
        self.assertMessageEqual(m, command='731', # RPL_MONOFFLINE
                fail_msg='Did not reply with 731 (RPL_MONOFFLINE) to '
                '“MONITOR + {}”, while “{}” is offline: {msg}',
                extra_format=(nick, nick))
        self.assertEqual(len(m.params), 2, m,
                fail_msg='Invalid number of params of RPL_MONOFFLINE: {msg}')
        self.assertEqual(m.params[1].split('!')[0], 'bar',
                fail_msg='731 (RPL_MONOFFLINE) reply to “MONITOR + {}” '
                'with bad target: {msg}',
                extra_format=(nick,))

    @cases.SpecificationSelector.requiredBySpecification('IRCv3.2')
    def testMonitorOneDisconnected(self):
        """“If any of the targets being added are online, the server will
        generate RPL_MONONLINE numerics listing those targets that are
        online.”
        -- <http://ircv3.net/specs/core/monitor-3.2.html#monitor--targettarget2>
        """
        self.connectClient('foo')
        self.check_server_support()
        self.sendLine(1, 'MONITOR + bar')
        self.assertMonoffline(1, 'bar')
        self.connectClient('bar')
        self.assertMononline(1, 'bar')
        self.sendLine(2, 'QUIT :bye')
        try:
            self.getMessages(2)
        except ConnectionResetError:
            pass
        self.assertMonoffline(1, 'bar')

    @cases.SpecificationSelector.requiredBySpecification('IRCv3.2')
    def testMonitorOneConnection(self):
        self.connectClient('foo')
        self.check_server_support()
        self.sendLine(1, 'MONITOR + bar')
        self.getMessages(1)
        self.connectClient('bar')
        self.assertMononline(1, 'bar')

    @cases.SpecificationSelector.requiredBySpecification('IRCv3.2')
    def testMonitorOneConnected(self):
        """“If any of the targets being added are offline, the server will
        generate RPL_MONOFFLINE numerics listing those targets that are
        online.”
        -- <http://ircv3.net/specs/core/monitor-3.2.html#monitor--targettarget2>
        """
        self.connectClient('foo')
        self.check_server_support()
        self.connectClient('bar')
        self.sendLine(1, 'MONITOR + bar')
        self.assertMononline(1, 'bar')
        self.sendLine(2, 'QUIT :bye')
        try:
            self.getMessages(2)
        except ConnectionResetError:
            pass
        self.assertMonoffline(1, 'bar')

    @cases.SpecificationSelector.requiredBySpecification('IRCv3.2')
    def testMonitorOneConnection(self):
        self.connectClient('foo')
        self.check_server_support()
        self.connectClient('bar')
        self.sendLine(1, 'MONITOR + bar')
        self.assertMononline(1, 'bar')
        self.sendLine(2, 'QUIT :bye')
        try:
            self.getMessages(2)
        except ConnectionResetError:
            pass
        self.assertMonoffline(1, 'bar')
        self.connectClient('bar')
        self.assertMononline(1, 'bar')

    @cases.SpecificationSelector.requiredBySpecification('IRCv3.2')
    def testMonitorConnectedAndDisconnected(self):
        """“If any of the targets being added are online, the server will
        generate RPL_MONONLINE numerics listing those targets that are
        online.

        If any of the targets being added are offline, the server will
        generate RPL_MONOFFLINE numerics listing those targets that are
        online.”
        -- <http://ircv3.net/specs/core/monitor-3.2.html#monitor--targettarget2>
        """
        self.connectClient('foo')
        self.check_server_support()
        self.connectClient('bar')
        self.sendLine(1, 'MONITOR + bar,baz')
        m1 = self.getMessage(1)
        m2 = self.getMessage(1)
        commands = {m1.command, m2.command}
        self.assertEqual(commands, {'730', '731'},
                fail_msg='Did not send one 730 (RPL_MONONLINE) and one '
                '731 (RPL_MONOFFLINE) after “MONITOR + bar,baz” when “bar” '
                'is online and “baz” is offline. Sent this instead: {}',
                extra_format=((m1, m2)))
        if m1.command == '731':
            (m1, m2) = (m2, m1)
        self.assertEqual(len(m1.params), 2, m1,
                fail_msg='Invalid number of params of RPL_MONONLINE: {msg}')
        self.assertEqual(len(m2.params), 2, m2,
                fail_msg='Invalid number of params of RPL_MONONLINE: {msg}')
        self.assertEqual(m1.params[1].split('!')[0], 'bar', m1,
                fail_msg='730 (RPL_MONONLINE) with bad target after '
                '“MONITOR + bar,baz” and “bar” is connected: {msg}')
        self.assertEqual(m2.params[1].split('!')[0], 'baz', m2,
                fail_msg='731 (RPL_MONOFFLINE) with bad target after '
                '“MONITOR + bar,baz” and “baz” is disconnected: {msg}')

    @cases.SpecificationSelector.requiredBySpecification('IRCv3.2')
    def testUnmonitor(self):
        self.connectClient('foo')
        self.check_server_support()
        self.sendLine(1, 'MONITOR + bar')
        self.getMessages(1)
        self.connectClient('bar')
        self.assertMononline(1, 'bar')
        self.sendLine(1, 'MONITOR - bar')
        self.assertEqual(self.getMessages(1), [],
                fail_msg='Got messages after “MONITOR - bar”: {got}')
        self.sendLine(2, 'QUIT :bye')
        try:
            self.getMessages(2)
        except ConnectionResetError:
            pass
        self.assertEqual(self.getMessages(1), [],
                fail_msg='Got messages after disconnection of unmonitored '
                'nick: {got}')

    @cases.SpecificationSelector.requiredBySpecification('IRCv3.2')
    def testMonitorForbidsMasks(self):
        """“The MONITOR implementation also enhances user privacy by
        disallowing subscription to hostmasks, allowing users to avoid
        nick-change stalking.”
        -- <http://ircv3.net/specs/core/monitor-3.2.html#watch-vs-monitor>

        “For this specification, ‘target’ MUST be a valid nick as determined
        by the IRC daemon.”
        -- <http://ircv3.net/specs/core/monitor-3.2.html#monitor-command>
        """
        self.connectClient('foo')
        self.check_server_support()
        self.sendLine(1, 'MONITOR + *!username@localhost')
        self.sendLine(1, 'MONITOR + *!username@127.0.0.1')
        try:
            m = self.getMessage(1)
            self.assertNotEqual(m.command, '731', m,
                    fail_msg='Got 731 (RPL_MONOFFLINE) after adding a monitor '
                    'on a mask: {msg}')
        except NoMessageException:
            pass
        self.connectClient('bar')
        try:
            m = self.getMessage(1)
        except NoMessageException:
            pass
        else:
            raise AssertionError('Got message after client whose MONITORing '
                    'was requested via hostmask connected: {}'.format(m))

    @cases.SpecificationSelector.requiredBySpecification('IRCv3.2')
    def testTwoMonitoringOneRemove(self):
        """Tests the following scenario:
        * foo MONITORs qux
        * bar MONITORs qux
        * bar unMONITORs qux
        * qux connects.
        """
        self.connectClient('foo')
        self.check_server_support()
        self.connectClient('bar')
        self.sendLine(1, 'MONITOR + qux')
        self.sendLine(2, 'MONITOR + qux')
        self.getMessages(1)
        self.getMessages(2)
        self.sendLine(2, 'MONITOR - qux')
        l = self.getMessages(2)
        self.assertEqual(l, [],
                fail_msg='Got response to “MONITOR -”: {}',
                extra_format=(l,))
        self.connectClient('qux')
        self.getMessages(3)
        l = self.getMessages(1)
        self.assertNotEqual(l, [],
                fail_msg='Received no message after MONITORed client '
                'connects.')
        l = self.getMessages(2)
        self.assertEqual(l, [],
                fail_msg='Got response to unmonitored client: {}',
                extra_format=(l,))



import re
import collections
import supybot.utils

# http://ircv3.net/specs/core/message-tags-3.2.html#escaping-values
TAG_ESCAPE = [
    ('\\', '\\\\'), # \ -> \\
    (' ', r'\s'),
    (';', r'\:'),
    ('\r', r'\r'),
    ('\n', r'\n'),
    ]
unescape_tag_value = supybot.utils.str.MultipleReplacer(
        dict(map(lambda x:(x[1],x[0]), TAG_ESCAPE)))

# TODO: validate host
tag_key_validator = re.compile('(\S+/)?[a-zA-Z0-9-]+')

def parse_tags(s):
    tags = {}
    for tag in s.split(';'):
        if '=' not in tag:
            tags[tag] = None
        else:
            (key, value) = tag.split('=', 1)
            assert tag_key_validator.match(key), \
                    'Invalid tag key: {}'.format(key)
            tags[key] = unescape_tag_value(value)
    return tags

Message = collections.namedtuple('Message',
        'tags prefix command params')

def parse_message(s):
    """Parse a message according to
    http://tools.ietf.org/html/rfc1459#section-2.3.1
    and
    http://ircv3.net/specs/core/message-tags-3.2.html"""
    assert s.endswith('\r\n'), 'Message does not end with CR LF: {!r}'.format(s)
    s = s[0:-2]
    if s.startswith('@'):
        (tags, s) = s.split(' ', 1)
        tags = parse_tags(tags[1:])
    else:
        tags = []
    if ' :' in s:
        (other_tokens, trailing_param) = s.split(' :', 1)
        tokens = list(filter(bool, other_tokens.split(' '))) + [trailing_param]
    else:
        tokens = list(filter(bool, s.split(' ')))
    if tokens[0].startswith(':'):
        prefix = tokens.pop(0)[1:]
    else:
        prefix = None
    command = tokens.pop(0)
    params = tokens
    return Message(
            tags=tags,
            prefix=prefix,
            command=command,
            params=params,
            )



"""
Handles ambiguities of RFCs.
"""

def normalize_namreply_params(params):
    # So… RFC 2812 says:
    #       "( "=" / "*" / "@" ) <channel>
    #       :[ "@" / "+" ] <nick> *( " " [ "@" / "+" ] <nick> )
    # but spaces seem to be missing (eg. before the colon), so we
    # don't know if there should be one before the <channel> and its
    # prefix.
    # So let's normalize this to “with space”, and strip spaces at the
    # end of the nick list.
    if len(params) == 3:
        assert params[1][0] in '=*@', params
        params.insert(1), params[1][0]
        params[2] = params[2][1:]
    params[3] = params[3].rstrip()
    return params



import enum
import collections

@enum.unique
class Mechanisms(enum.Enum):
    """Enumeration for representing possible mechanisms."""
    @classmethod
    def as_string(cls, mech):
        return {cls.plain: 'PLAIN',
                cls.ecdsa_nist256p_challenge: 'ECDSA-NIST256P-CHALLENGE',
                cls.scram_sha_256: 'SCRAM-SHA-256',
                }[mech]
    plain = 1
    ecdsa_nist256p_challenge = 2
    scram_sha_256 = 3

Authentication = collections.namedtuple('Authentication',
        'mechanisms username password ecdsa_key')
Authentication.__new__.__defaults__ = ([Mechanisms.plain], None, None, None)



import base64

from irctest import cases
from irctest.irc_utils.message_parser import Message

class RegistrationTestCase(cases.BaseServerTestCase):
    def testRegistration(self):
        self.controller.registerUser(self, 'testuser', 'mypassword')

class SaslTestCase(cases.BaseServerTestCase, cases.OptionalityHelper):
    @cases.SpecificationSelector.requiredBySpecification('IRCv3.1')
    @cases.OptionalityHelper.skipUnlessHasMechanism('PLAIN')
    def testPlain(self):
        """PLAIN authentication with correct username/password."""
        self.controller.registerUser(self, 'foo', 'sesame')
        self.controller.registerUser(self, 'jilles', 'sesame')
        self.controller.registerUser(self, 'bar', 'sesame')
        self.addClient()
        self.sendLine(1, 'CAP LS 302')
        capabilities = self.getCapLs(1)
        self.assertIn('sasl', capabilities,
                fail_msg='Does not have SASL as the controller claims.')
        if capabilities['sasl'] is not None:
            self.assertIn('PLAIN', capabilities['sasl'],
                fail_msg='Does not have PLAIN mechanism as the controller '
                'claims')
        self.sendLine(1, 'AUTHENTICATE PLAIN')
        m = self.getMessage(1, filter_pred=lambda m:m.command != 'NOTICE')
        self.assertMessageEqual(m, command='AUTHENTICATE', params=['+'],
                fail_msg='Sent “AUTHENTICATE PLAIN”, server should have '
                'replied with “AUTHENTICATE +”, but instead sent: {msg}')
        self.sendLine(1, 'AUTHENTICATE amlsbGVzAGppbGxlcwBzZXNhbWU=')
        m = self.getMessage(1, filter_pred=lambda m:m.command != 'NOTICE')
        self.assertMessageEqual(m, command='900',
                fail_msg='Did not send 900 after correct SASL authentication.')
        self.assertEqual(m.params[2], 'jilles', m,
                fail_msg='900 should contain the account name as 3rd argument '
                '({expects}), not {got}: {msg}')

    @cases.SpecificationSelector.requiredBySpecification('IRCv3.1')
    @cases.OptionalityHelper.skipUnlessHasMechanism('PLAIN')
    def testPlainNoAuthzid(self):
        """“message   = [authzid] UTF8NUL authcid UTF8NUL passwd

        […]

        Upon receipt of the message, the server will verify the presented (in
        the message) authentication identity (authcid) and password (passwd)
        with the system authentication database, and it will verify that the
        authentication credentials permit the client to act as the (presented
        or derived) authorization identity (authzid).  If both steps succeed,
        the user is authenticated.

        […]


        When no authorization identity is provided, the server derives an
        authorization identity from the prepared representation of the
        provided authentication identity string.  This ensures that the
        derivation of different representations of the authentication
        identity produces the same authorization identity.”
        -- <https://tools.ietf.org/html/rfc4616#section-2>
        """
        self.controller.registerUser(self, 'foo', 'sesame')
        self.controller.registerUser(self, 'jilles', 'sesame')
        self.controller.registerUser(self, 'bar', 'sesame')
        self.addClient()
        self.sendLine(1, 'CAP LS 302')
        capabilities = self.getCapLs(1)
        self.assertIn('sasl', capabilities,
                fail_msg='Does not have SASL as the controller claims.')
        if capabilities['sasl'] is not None:
            self.assertIn('PLAIN', capabilities['sasl'],
                fail_msg='Does not have PLAIN mechanism as the controller '
                'claims')
        self.sendLine(1, 'AUTHENTICATE PLAIN')
        m = self.getMessage(1, filter_pred=lambda m:m.command != 'NOTICE')
        self.assertMessageEqual(m, command='AUTHENTICATE', params=['+'],
                fail_msg='Sent “AUTHENTICATE PLAIN”, server should have '
                'replied with “AUTHENTICATE +”, but instead sent: {msg}')
        self.sendLine(1, 'AUTHENTICATE AGppbGxlcwBzZXNhbWU=')
        m = self.getMessage(1, filter_pred=lambda m:m.command != 'NOTICE')
        self.assertMessageEqual(m, command='900',
                fail_msg='Did not send 900 after correct SASL authentication.')
        self.assertEqual(m.params[2], 'jilles', m,
                fail_msg='900 should contain the account name as 3rd argument '
                '({expects}), not {got}: {msg}')

    @cases.SpecificationSelector.requiredBySpecification('IRCv3.1')
    def testMechanismNotAvailable(self):
        """“If authentication fails, a 904 or 905 numeric will be sent”
        -- <http://ircv3.net/specs/extensions/sasl-3.1.html#the-authenticate-command>
        """
        self.controller.registerUser(self, 'jilles', 'sesame')
        self.addClient()
        self.sendLine(1, 'CAP LS 302')
        capabilities = self.getCapLs(1)
        self.assertIn('sasl', capabilities,
                fail_msg='Does not have SASL as the controller claims.')
        self.sendLine(1, 'AUTHENTICATE FOO')
        m = self.getRegistrationMessage(1)
        self.assertMessageEqual(m, command='904',
                fail_msg='Did not reply with 904 to “AUTHENTICATE FOO”: {msg}')

    @cases.SpecificationSelector.requiredBySpecification('IRCv3.1')
    @cases.OptionalityHelper.skipUnlessHasMechanism('PLAIN')
    def testPlainLarge(self):
        """Test the client splits large AUTHENTICATE messages whose payload
        is not a multiple of 400.
        <http://ircv3.net/specs/extensions/sasl-3.1.html#the-authenticate-command>
        """
        self.controller.registerUser(self, 'foo', 'bar'*100)
        authstring = base64.b64encode(b'\x00'.join(
            [b'foo', b'foo', b'bar'*100])).decode()
        self.addClient()
        self.sendLine(1, 'CAP LS 302')
        capabilities = self.getCapLs(1)
        self.assertIn('sasl', capabilities,
                fail_msg='Does not have SASL as the controller claims.')
        if capabilities['sasl'] is not None:
            self.assertIn('PLAIN', capabilities['sasl'],
                fail_msg='Does not have PLAIN mechanism as the controller '
                'claims')
        self.sendLine(1, 'AUTHENTICATE PLAIN')
        m = self.getRegistrationMessage(1)
        self.assertMessageEqual(m, command='AUTHENTICATE', params=['+'],
                fail_msg='Sent “AUTHENTICATE PLAIN”, expected '
                '“AUTHENTICATE +” as a response, but got: {msg}')
        self.sendLine(1, 'AUTHENTICATE {}'.format(authstring[0:400]))
        self.sendLine(1, 'AUTHENTICATE {}'.format(authstring[400:]))

        # TODO: may be in the other order
        m = self.getRegistrationMessage(1)
        self.assertMessageEqual(m, command='900',
                fail_msg='Expected 900 (RPL_LOGGEDIN) after successful '
                'login, but got: {msg}')
        m = self.getRegistrationMessage(1)
        self.assertMessageEqual(m, command='903',
                fail_msg='Expected 900 (RPL_LOGGEDIN) after successful '
                'login, but got: {msg}')

    # TODO: add a test for when the length of the authstring is greater than 800.
    # I don't know how to do it, because it would make the registration
    # message's length too big for it to be valid.

    @cases.SpecificationSelector.requiredBySpecification('IRCv3.1')
    @cases.OptionalityHelper.skipUnlessHasMechanism('PLAIN')
    def testPlainLargeEquals400(self):
        """Test the client splits large AUTHENTICATE messages whose payload
        is not a multiple of 400.
        <http://ircv3.net/specs/extensions/sasl-3.1.html#the-authenticate-command>
        """
        self.controller.registerUser(self, 'foo', 'bar'*97)
        authstring = base64.b64encode(b'\x00'.join(
            [b'foo', b'foo', b'bar'*97])).decode()
        assert len(authstring) == 400, 'Bad test'
        self.addClient()
        self.sendLine(1, 'CAP LS 302')
        capabilities = self.getCapLs(1)
        self.assertIn('sasl', capabilities,
                fail_msg='Does not have SASL as the controller claims.')
        if capabilities['sasl'] is not None:
            self.assertIn('PLAIN', capabilities['sasl'],
                fail_msg='Does not have PLAIN mechanism as the controller '
                'claims')
        self.sendLine(1, 'AUTHENTICATE PLAIN')
        m = self.getRegistrationMessage(1)
        self.assertMessageEqual(m, command='AUTHENTICATE', params=['+'],
                fail_msg='Sent “AUTHENTICATE PLAIN”, expected '
                '“AUTHENTICATE +” as a response, but got: {msg}')
        self.sendLine(1, 'AUTHENTICATE {}'.format(authstring))
        self.sendLine(1, 'AUTHENTICATE +')

        # TODO: may be in the other order
        m = self.getRegistrationMessage(1)
        self.assertMessageEqual(m, command='900',
                fail_msg='Expected 900 (RPL_LOGGEDIN) after successful '
                'login, but got: {msg}')
        m = self.getRegistrationMessage(1)
        self.assertMessageEqual(m, command='903',
                fail_msg='Expected 900 (RPL_LOGGEDIN) after successful '
                'login, but got: {msg}')

    # TODO: add a test for when the length of the authstring is 800.
    # I don't know how to do it, because it would make the registration
    # message's length too big for it to be valid.



import ssl
import time
import socket
from .irc_utils import message_parser
from .exceptions import NoMessageException, ConnectionClosed

class ClientMock:
    def __init__(self, name, show_io):
        self.name = name
        self.show_io = show_io
        self.inbuffer = []
        self.ssl = False
    def connect(self, hostname, port):
        self.conn = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.conn.settimeout(1) # TODO: configurable
        self.conn.connect((hostname, port))
        if self.show_io:
            print('{:.3f} {}: connects to server.'.format(time.time(), self.name))
    def disconnect(self):
        if self.show_io:
            print('{:.3f} {}: disconnects from server.'.format(time.time(), self.name))
        self.conn.close()
    def starttls(self):
        assert not self.ssl, 'SSL already active.'
        self.conn = ssl.wrap_socket(self.conn)
        self.ssl = True
    def getMessages(self, synchronize=True, assert_get_one=False):
        if synchronize:
            token = 'synchronize{}'.format(time.monotonic())
            self.sendLine('PING {}'.format(token))
        got_pong = False
        data = b''
        (self.inbuffer, messages) = ([], self.inbuffer)
        conn = self.conn
        try:
            while not got_pong:
                try:
                    new_data = conn.recv(4096)
                except socket.timeout:
                    if not assert_get_one and not synchronize and data == b'':
                        # Received nothing
                        return []
                    if self.show_io:
                        print('{:.3f} waiting…'.format(time.time()))
                    time.sleep(0.1)
                    continue
                except ConnectionResetError:
                    raise ConnectionClosed()
                else:
                    if not new_data:
                        # Connection closed
                        raise ConnectionClosed()
                data += new_data
                if not new_data.endswith(b'\r\n'):
                    time.sleep(0.1)
                    continue
                if not synchronize:
                    got_pong = True
                for line in data.decode().split('\r\n'):
                    if line:
                        if self.show_io:
                            print('{time:.3f}{ssl} S -> {client}: {line}'.format(
                                time=time.time(),
                                ssl=' (ssl)' if self.ssl else '',
                                client=self.name,
                                line=line))
                        message = message_parser.parse_message(line + '\r\n')
                        if message.command == 'PONG' and \
                                token in message.params:
                            got_pong = True
                        else:
                            messages.append(message)
                data = b''
        except ConnectionClosed:
            if messages:
                return messages
            else:
                raise
        else:
            return messages
    def getMessage(self, filter_pred=None, synchronize=True):
        while True:
            if not self.inbuffer:
                self.inbuffer = self.getMessages(
                        synchronize=synchronize, assert_get_one=True)
            if not self.inbuffer:
                raise NoMessageException()
            message = self.inbuffer.pop(0) # TODO: use dequeue
            if not filter_pred or filter_pred(message):
                return message
    def sendLine(self, line):
        if not line.endswith('\r\n'):
            line += '\r\n'
        encoded_line = line.encode()
        try:
            ret = self.conn.sendall(encoded_line)
        except BrokenPipeError:
            raise ConnectionClosed()
        if self.ssl: # https://bugs.python.org/issue25951
            assert ret == len(encoded_line), (ret, repr(encoded_line))
        else:
            assert ret is None, ret
        if self.show_io:
            print('{time:.3f}{ssl} {client} -> S: {line}'.format(
                time=time.time(),
                ssl=' (ssl)' if self.ssl else '',
                client=self.name,
                line=line.strip('\r\n')))



from irctest import cases
from irctest.irc_utils.message_parser import Message

class CapTestCase(cases.BaseServerTestCase):
    @cases.SpecificationSelector.requiredBySpecification('IRCv3.1')
    def testNoReq(self):
        """Test the server handles gracefully clients which do not send
        REQs.

        “Clients that support capabilities but do not wish to enter
        negotiation SHOULD send CAP END upon connection to the server.”
        -- <http://ircv3.net/specs/core/capability-negotiation-3.1.html#the-cap-end-subcommand>
        """
        self.addClient(1)
        self.sendLine(1, 'CAP LS 302')
        self.getCapLs(1)
        self.sendLine(1, 'USER foo foo foo :foo')
        self.sendLine(1, 'NICK foo')
        self.sendLine(1, 'CAP END')
        m = self.getRegistrationMessage(1)
        self.assertMessageEqual(m, command='001',
                fail_msg='Expected 001 after sending CAP END, got {msg}.')

    @cases.SpecificationSelector.requiredBySpecification('IRCv3.1')
    def testReqUnavailable(self):
        """Test the server handles gracefully clients which request
        capabilities that are not available.
        <http://ircv3.net/specs/core/capability-negotiation-3.1.html>
        """
        self.addClient(1)
        self.sendLine(1, 'CAP LS 302')
        self.getCapLs(1)
        self.sendLine(1, 'USER foo foo foo :foo')
        self.sendLine(1, 'NICK foo')
        self.sendLine(1, 'CAP REQ :foo')
        m = self.getRegistrationMessage(1)
        self.assertMessageEqual(m, command='CAP',
                subcommand='NAK', subparams=['foo'],
                fail_msg='Expected CAP NAK after requesting non-existing '
                'capability, got {msg}.')
        self.sendLine(1, 'CAP END')
        m = self.getRegistrationMessage(1)
        self.assertMessageEqual(m, command='001',
                fail_msg='Expected 001 after sending CAP END, got {msg}.')

    @cases.SpecificationSelector.requiredBySpecification('IRCv3.1')
    def testNakExactString(self):
        """“The argument of the NAK subcommand MUST consist of at least the
        first 100 characters of the capability list in the REQ subcommand which
        triggered the NAK.”
        -- <http://ircv3.net/specs/core/capability-negotiation-3.1.html#the-cap-nak-subcommand>
        """
        self.addClient(1)
        self.sendLine(1, 'CAP LS 302')
        self.getCapLs(1)
        # Five should be enough to check there is no reordering, even
        # alphabetical
        self.sendLine(1, 'CAP REQ :foo qux bar baz qux quux')
        m = self.getRegistrationMessage(1)
        self.assertMessageEqual(m, command='CAP',
                subcommand='NAK', subparams=['foo qux bar baz qux quux'],
                fail_msg='Expected “CAP NAK :foo qux bar baz qux quux” after '
                'sending “CAP REQ :foo qux bar baz qux quux”, but got {msg}.')

    @cases.SpecificationSelector.requiredBySpecification('IRCv3.1')
    def testNakWhole(self):
        """“The capability identifier set must be accepted as a whole, or
        rejected entirely.”
        -- <http://ircv3.net/specs/core/capability-negotiation-3.1.html#the-cap-req-subcommand>
        """
        self.addClient(1)
        self.sendLine(1, 'CAP LS 302')
        self.assertIn('multi-prefix', self.getCapLs(1))
        self.sendLine(1, 'CAP REQ :foo multi-prefix bar')
        m = self.getRegistrationMessage(1)
        self.assertMessageEqual(m, command='CAP',
                subcommand='NAK', subparams=['foo multi-prefix bar'],
                fail_msg='Expected “CAP NAK :foo multi-prefix bar” after '
                'sending “CAP REQ :foo multi-prefix bar”, but got {msg}.')
        self.sendLine(1, 'CAP REQ :multi-prefix bar')
        m = self.getRegistrationMessage(1)
        self.assertMessageEqual(m, command='CAP',
                subcommand='NAK', subparams=['multi-prefix bar'],
                fail_msg='Expected “CAP NAK :multi-prefix bar” after '
                'sending “CAP REQ :multi-prefix bar”, but got {msg}.')
        self.sendLine(1, 'CAP REQ :foo multi-prefix')
        m = self.getRegistrationMessage(1)
        self.assertMessageEqual(m, command='CAP',
                subcommand='NAK', subparams=['foo multi-prefix'],
                fail_msg='Expected “CAP NAK :foo multi-prefix” after '
                'sending “CAP REQ :foo multi-prefix”, but got {msg}.')
        # TODO: make sure multi-prefix is not enabled at this point
        self.sendLine(1, 'CAP REQ :multi-prefix')
        m = self.getRegistrationMessage(1)
        self.assertMessageEqual(m, command='CAP',
                subcommand='ACK', subparams=['multi-prefix'],
                fail_msg='Expected “CAP ACK :multi-prefix” after '
                'sending “CAP REQ :multi-prefix”, but got {msg}.')



"""
<http://ircv3.net/specs/extensions/extended-join-3.1.html>
"""

from irctest import cases
from irctest.irc_utils.message_parser import Message

class MetadataTestCase(cases.BaseServerTestCase, cases.OptionalityHelper):
    def connectRegisteredClient(self, nick):
        self.addClient()
        self.sendLine(2, 'CAP LS 302')
        capabilities = self.getCapLs(2)
        assert 'sasl' in capabilities
        self.sendLine(2, 'AUTHENTICATE PLAIN')
        m = self.getMessage(2, filter_pred=lambda m:m.command != 'NOTICE')
        self.assertMessageEqual(m, command='AUTHENTICATE', params=['+'],
                fail_msg='Sent “AUTHENTICATE PLAIN”, server should have '
                'replied with “AUTHENTICATE +”, but instead sent: {msg}')
        self.sendLine(2, 'AUTHENTICATE amlsbGVzAGppbGxlcwBzZXNhbWU=')
        m = self.getMessage(2, filter_pred=lambda m:m.command != 'NOTICE')
        self.assertMessageEqual(m, command='900',
                fail_msg='Did not send 900 after correct SASL authentication.')
        self.sendLine(2, 'USER f * * :Realname')
        self.sendLine(2, 'NICK {}'.format(nick))
        self.sendLine(2, 'CAP END')
        self.skipToWelcome(2)

    @cases.SpecificationSelector.requiredBySpecification('IRCv3.1')
    def testNotLoggedIn(self):
        self.connectClient('foo', capabilities=['extended-join'],
                skip_if_cap_nak=True)
        self.sendLine(1, 'JOIN #chan')
        self.getMessages(1)
        self.connectClient('bar')
        self.sendLine(2, 'JOIN #chan')
        m = self.getMessage(1)
        self.assertMessageEqual(m, command='JOIN',
                params=['#chan', '*', 'Realname'],
                fail_msg='Expected “JOIN #chan * :Realname” after '
                'unregistered user joined, got: {msg}')

    @cases.SpecificationSelector.requiredBySpecification('IRCv3.1')
    @cases.OptionalityHelper.skipUnlessHasMechanism('PLAIN')
    def testNotLoggedIn(self):
        self.connectClient('foo', capabilities=['extended-join'],
                skip_if_cap_nak=True)
        self.sendLine(1, 'JOIN #chan')
        self.getMessages(1)

        self.controller.registerUser(self, 'jilles', 'sesame')
        self.connectRegisteredClient('bar')

        self.sendLine(2, 'JOIN #chan')
        m = self.getMessage(1)
        self.assertMessageEqual(m, command='JOIN',
                params=['#chan', 'jilles', 'Realname'],
                fail_msg='Expected “JOIN #chan * :Realname” after '
                'nick “bar” logged in as “jilles” joined, got: {msg}')



def cap_list_to_dict(l):
    d = {}
    for cap in l:
        if '=' in cap:
            (key, value) = cap.split('=', 1)
        else:
            key = cap
            value = None
        d[key] = value
    return d



#!/usr/bin/env python3

import os
import sys
from setuptools import setup

if sys.version_info < (3, 4, 0):
    sys.stderr.write("This script requires Python 3.4 or newer.")
    sys.stderr.write(os.linesep)
    sys.exit(-1)

with open(os.path.join(os.path.dirname(__file__), 'requirements.txt')) as fd:
    requirements = [x.split('#')[0]
                    for x in fd.readlines()]

setup(
    name='irctest',
    version='0.1.1',
    author='Valentin Lorentz',
    url='https://github.com/ProgVal/irctest/',
    author_email='progval+irctest@progval.net',
    description='A script to test interoperability of IRC software.',
    platforms=['linux', 'linux2'],
    long_description="""This script aims at testing interoperability of
    software using the IRC protocol, by running them against test suites
    and making different software communicate with each other.""",
    classifiers = [
        'Development Status :: 2 - Pre-Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Communications :: Chat :: Internet Relay Chat',
        'Topic :: Software Development :: Testing',
        ],

    # Installation data
    packages=[
            'irctest',
            'irctest.client_tests',
            'irctest.controllers',
            'irctest.irc_utils',
            'irctest.server_tests',
            ],
    install_requires=requirements,
    )

# vim:set shiftwidth=4 softtabstop=4 expandtab textwidth=79:



"""
Section 3.2 of RFC 2812
<https://tools.ietf.org/html/rfc2812#section-3.2>
"""

from irctest import cases
from irctest import client_mock
from irctest import runner
from irctest.irc_utils import ambiguities
from irctest.irc_utils.message_parser import Message

class JoinTestCase(cases.BaseServerTestCase):
    @cases.SpecificationSelector.requiredBySpecification('RFC1459', 'RFC2812',
            strict=True)
    def testJoinAllMessages(self):
        """“If a JOIN is successful, the user receives a JOIN message as
        confirmation and is then sent the channel's topic (using RPL_TOPIC) and
        the list of users who are on the channel (using RPL_NAMREPLY), which
        MUST include the user joining.”
        -- <https://tools.ietf.org/html/rfc2812#section-3.2.1>

        “If a JOIN is successful, the user is then sent the channel's topic
        (using RPL_TOPIC) and the list of users who are on the channel (using
        RPL_NAMREPLY), which must include the user joining.”
        -- <https://tools.ietf.org/html/rfc1459#section-4.2.1>
        """
        self.connectClient('foo')
        self.sendLine(1, 'JOIN #chan')
        received_commands = {m.command for m in self.getMessages(1)}
        expected_commands = {
                '353', # RPL_NAMREPLY
                '366', # RPL_ENDOFNAMES
                }
        self.assertTrue(expected_commands.issubset(received_commands),
                'Server sent {} commands, but at least {} were expected.'
                .format(received_commands, expected_commands))
        self.assertTrue(received_commands & {'331', '332'} != set(), # RPL_NOTOPIC, RPL_TOPIC
                'Server sent neither 331 (RPL_NOTOPIC) or 332 (RPL_TOPIC)')

    @cases.SpecificationSelector.requiredBySpecification('RFC2812')
    def testJoinNamreply(self):
        """“353    RPL_NAMREPLY
            "( "=" / "*" / "@" ) <channel>
             :[ "@" / "+" ] <nick> *( " " [ "@" / "+" ] <nick> )”
        -- <https://tools.ietf.org/html/rfc2812#section-5.2>

        This test makes a user join and check what is sent to them.
        """
        self.connectClient('foo')
        self.sendLine(1, 'JOIN #chan')

        for m in self.getMessages(1):
            if m.command == '353':
                self.assertIn(len(m.params), (3, 4), m,
                        fail_msg='RPL_NAM_REPLY with number of arguments '
                        '<3 or >4: {msg}')
                params = ambiguities.normalize_namreply_params(m.params)
                self.assertIn(params[1], '=*@', m,
                        fail_msg='Bad channel prefix: {item} not in {list}: {msg}')
                self.assertEqual(params[2], '#chan', m,
                        fail_msg='Bad channel name: {got} instead of '
                        '{expects}: {msg}')
                self.assertIn(params[3], {'foo', '@foo', '+foo'}, m,
                        fail_msg='Bad user list: should contain only user '
                        '"foo" with an optional "+" or "@" prefix, but got: '
                        '{msg}')


    @cases.SpecificationSelector.requiredBySpecification('RFC1459', 'RFC2812')
    def testPartNotInEmptyChannel(self):
        """“442     ERR_NOTONCHANNEL
            "<channel> :You're not on that channel"

        - Returned by the server whenever a client tries to
          perform a channel effecting command for which the
          client isn't a member.”
          -- <https://tools.ietf.org/html/rfc1459#section-6.1>
          and <https://tools.ietf.org/html/rfc2812#section-5.2>

        According to RFCs, ERR_NOSUCHCHANNEL should only be used for invalid
        channel names:
        “403     ERR_NOSUCHCHANNEL
          "<channel name> :No such channel"

        - Used to indicate the given channel name is invalid.”
        -- <https://tools.ietf.org/html/rfc1459#section-6.1>
        and <https://tools.ietf.org/html/rfc2812#section-5.2>

        However, many implementations use 479 instead, so let's allow it.
        <http://danieloaks.net/irc-defs/defs/ircnumerics.html#403>
        <http://danieloaks.net/irc-defs/defs/ircnumerics.html#479>
        """
        self.connectClient('foo')
        self.sendLine(1, 'PART #chan')
        m = self.getMessage(1)
        self.assertIn(m.command, {'442', '403'}, m, # ERR_NOTONCHANNEL, ERR_NOSUCHCHANNEL
                fail_msg='Expected ERR_NOTONCHANNEL (442) or '
                'ERR_NOSUCHCHANNEL (403) after PARTing an empty channel '
                'one is not on, but got: {msg}')

    @cases.SpecificationSelector.requiredBySpecification('RFC1459', 'RFC2812')
    def testPartNotInNonEmptyChannel(self):
        self.connectClient('foo')
        self.connectClient('bar')
        self.sendLine(1, 'JOIN #chan')
        self.getMessages(1) # Synchronize
        self.sendLine(2, 'PART #chan')
        m = self.getMessage(2)
        self.assertMessageEqual(m, command='442', # ERR_NOTONCHANNEL
                fail_msg='Expected ERR_NOTONCHANNEL (442) '
                'after PARTing a non-empty channel '
                'one is not on, but got: {msg}')
        self.assertEqual(self.getMessages(2), [])
    testPartNotInNonEmptyChannel.__doc__ = testPartNotInEmptyChannel.__doc__

    def testJoinTwice(self):
        self.connectClient('foo')
        self.sendLine(1, 'JOIN #chan')
        m = self.getMessage(1)
        self.assertMessageEqual(m, command='JOIN', params=['#chan'])
        self.getMessages(1)
        self.sendLine(1, 'JOIN #chan')
        # Note that there may be no message. Both RFCs require replies only
        # if the join is successful, or has an error among the given set.
        for m in self.getMessages(1):
            if m.command == '353':
                self.assertIn(len(m.params), (3, 4), m,
                        fail_msg='RPL_NAM_REPLY with number of arguments '
                        '<3 or >4: {msg}')
                params = ambiguities.normalize_namreply_params(m.params)
                self.assertIn(params[1], '=*@', m,
                        fail_msg='Bad channel prefix: {item} not in {list}: {msg}')
                self.assertEqual(params[2], '#chan', m,
                        fail_msg='Bad channel name: {got} instead of '
                        '{expects}: {msg}')
                self.assertIn(params[3], {'foo', '@foo', '+foo'}, m,
                        fail_msg='Bad user list after user "foo" joined twice '
                        'the same channel: should contain only user '
                        '"foo" with an optional "+" or "@" prefix, but got: '
                        '{msg}')

    @cases.SpecificationSelector.requiredBySpecification('RFC1459', 'RFC2812')
    def testTopic(self):
        """“Once a user has joined a channel, he receives information about
        all commands his server receives affecting the channel.  This
        includes […] TOPIC”
        -- <https://tools.ietf.org/html/rfc1459#section-4.2.1>
        and <https://tools.ietf.org/html/rfc2812#section-3.2.1>
        """
        self.connectClient('foo')
        self.connectClient('bar')
        self.sendLine(1, 'JOIN #chan')
        self.sendLine(2, 'JOIN #chan')
        # TODO: check foo is opped OR +t is unset

        self.getMessages(1)
        self.getMessages(2)
        self.getMessages(1)

        self.sendLine(1, 'TOPIC #chan :T0P1C')
        try:
            m = self.getMessage(1)
            if m.command == '482':
                raise runner.ImplementationChoice(
                        'Channel creators are not opped by default, and '
                        'channel modes to no allow regular users to change '
                        'topic.')
            self.assertMessageEqual(m, command='TOPIC')
        except client_mock.NoMessageException:
            # The RFCs do not say TOPIC must be echoed
            pass
        m = self.getMessage(2)
        self.assertMessageEqual(m, command='TOPIC', params=['#chan', 'T0P1C'])

    @cases.SpecificationSelector.requiredBySpecification('RFC1459', 'RFC2812')
    def testTopicMode(self):
        """“Once a user has joined a channel, he receives information about
        all commands his server receives affecting the channel.  This
        includes […] TOPIC”
        -- <https://tools.ietf.org/html/rfc1459#section-4.2.1>
        and <https://tools.ietf.org/html/rfc2812#section-3.2.1>
        """
        self.connectClient('foo')
        self.connectClient('bar')
        self.sendLine(1, 'JOIN #chan')
        self.sendLine(2, 'JOIN #chan')
        # TODO: check foo is opped

        self.getMessages(1)
        self.getMessages(2)
        self.getMessages(1)

        self.sendLine(1, 'MODE #chan +t')
        try:
            m = self.getMessage(1)
            if m.command == '482':
                raise runner.ImplementationChoice(
                        'Channel creators are not opped by default.')
            self.assertMessageEqual(m, command='TOPIC')
        except client_mock.NoMessageException:
            # The RFCs do not say TOPIC must be echoed
            pass
        self.sendLine(2, 'TOPIC #chan :T0P1C')
        m = self.getMessage(2)
        self.assertMessageEqual(m, command='482',
                fail_msg='Non-op user was not refused use of TOPIC: {msg}')
        self.assertEqual(self.getMessages(1), [])

        self.sendLine(1, 'MODE #chan -t')
        self.getMessages(1)
        self.sendLine(2, 'TOPIC #chan :T0P1C')
        try:
            m = self.getMessage(2)
            self.assertNotEqual(m.command, '482',
                    msg='User was refused TOPIC whereas +t was not '
                    'set: {}'.format(m))
        except client_mock.NoMessageException:
            # The RFCs do not say TOPIC must be echoed
            pass
        m = self.getMessage(1)
        self.assertMessageEqual(m, command='TOPIC', params=['#chan', 'T0P1C'])



    @cases.SpecificationSelector.requiredBySpecification('RFC1459', 'RFC2812')
    def testListEmpty(self):
        """<https://tools.ietf.org/html/rfc1459#section-4.2.6>
        <https://tools.ietf.org/html/rfc2812#section-3.2.6>
        """
        self.connectClient('foo')
        self.connectClient('bar')
        self.getMessages(1)
        self.sendLine(2, 'LIST')
        m = self.getMessage(2)
        self.assertMessageEqual(m, command='321', # RPL_LISTSTART
                fail_msg='First reply to LIST is not 321 (RPL_LISTSTART), '
                'but: {msg}')
        m = self.getMessage(2)
        self.assertNotEqual(m.command, '322', # RPL_LIST
                'LIST response gives (at least) one channel, whereas there '
                'is none.')
        self.assertMessageEqual(m, command='323', # RPL_LISTEND
                fail_msg='Second reply to LIST is not 322 (RPL_LIST) '
                'or 323 (RPL_LISTEND), or but: {msg}')

    @cases.SpecificationSelector.requiredBySpecification('RFC1459', 'RFC2812')
    def testListOne(self):
        """When a channel exists, LIST should get it in a reply.
        <https://tools.ietf.org/html/rfc1459#section-4.2.6>
        <https://tools.ietf.org/html/rfc2812#section-3.2.6>
        """
        self.connectClient('foo')
        self.connectClient('bar')
        self.sendLine(1, 'JOIN #chan')
        self.getMessages(1)
        self.sendLine(2, 'LIST')
        m = self.getMessage(2)
        self.assertMessageEqual(m, command='321', # RPL_LISTSTART
                fail_msg='First reply to LIST is not 321 (RPL_LISTSTART), '
                'but: {msg}')
        m = self.getMessage(2)
        self.assertNotEqual(m.command, '323', # RPL_LISTEND
                fail_msg='LIST response ended (ie. 323, aka RPL_LISTEND) '
                'without listing any channel, whereas there is one.')
        self.assertMessageEqual(m, command='322', # RPL_LIST
                fail_msg='Second reply to LIST is not 322 (RPL_LIST), '
                'nor 323 (RPL_LISTEND) but: {msg}')
        m = self.getMessage(2)
        self.assertNotEqual(m.command, '322', # RPL_LIST
                fail_msg='LIST response gives (at least) two channels, '
                'whereas there is only one.')
        self.assertMessageEqual(m, command='323', # RPL_LISTEND
                fail_msg='Third reply to LIST is not 322 (RPL_LIST) '
                'or 323 (RPL_LISTEND), or but: {msg}')

    @cases.SpecificationSelector.requiredBySpecification('RFC1459', 'RFC2812')
    def testKickSendsMessages(self):
        """“Once a user has joined a channel, he receives information about
        all commands his server receives affecting the channel.  This
        includes […] KICK”
        -- <https://tools.ietf.org/html/rfc1459#section-4.2.1>
        and <https://tools.ietf.org/html/rfc2812#section-3.2.1>
        """
        self.connectClient('foo')
        self.connectClient('bar')
        self.connectClient('baz')
        self.sendLine(1, 'JOIN #chan')
        # TODO: check foo is an operator
        self.sendLine(2, 'JOIN #chan')
        self.sendLine(3, 'JOIN #chan')

        import time
        time.sleep(0.1)
        self.getMessages(1)
        self.getMessages(2)
        self.getMessages(3)
        self.sendLine(1, 'KICK #chan bar :bye')
        try:
            m = self.getMessage(1)
            if m.command == '482':
                raise runner.ImplementationChoice(
                        'Channel creators are not opped by default.')
            self.assertMessageEqual(m, command='KICK')
        except client_mock.NoMessageException:
            # The RFCs do not say KICK must be echoed
            pass
        m = self.getMessage(2)
        self.assertMessageEqual(m, command='KICK',
                params=['#chan', 'bar', 'bye'])
        m = self.getMessage(3)
        self.assertMessageEqual(m, command='KICK',
                params=['#chan', 'bar', 'bye'])

    @cases.SpecificationSelector.requiredBySpecification('RFC2812')
    def testDoubleKickMessages(self):
        """“The server MUST NOT send KICK messages with multiple channels or
        users to clients.  This is necessarily to maintain backward
        compatibility with old client software.”
        -- https://tools.ietf.org/html/rfc2812#section-3.2.8
        """
        self.connectClient('foo')
        self.connectClient('bar')
        self.connectClient('baz')
        self.connectClient('qux')
        self.sendLine(1, 'JOIN #chan')
        # TODO: check foo is an operator
        self.sendLine(2, 'JOIN #chan')
        self.sendLine(3, 'JOIN #chan')
        self.sendLine(4, 'JOIN #chan')

        # Synchronize
        self.getMessages(1)
        self.getMessages(2)
        self.getMessages(3)
        self.getMessages(4)
        self.getMessages(1)
        self.getMessages(2)
        self.getMessages(3)
        self.getMessages(4)

        self.sendLine(1, 'KICK #chan,#chan bar,baz :bye')
        try:
            m = self.getMessage(1)
            if m.command == '482':
                raise runner.OptionalExtensionNotSupported(
                        'Channel creators are not opped by default.')
            if m.command in {'401', '403'}:
                raise runner.NotImplementedByController(
                        'Multi-target KICK')
        except client_mock.NoMessageException:
            # The RFCs do not say KICK must be echoed
            pass

        # TODO: could be in the other order
        m = self.getMessage(4)
        self.assertMessageEqual(m, command='KICK',
                params=['#chan', 'bar', 'bye'])
        m = self.getMessage(4)
        self.assertMessageEqual(m, command='KICK',
                params=['#chan', 'baz', 'bye'])

    @cases.SpecificationSelector.requiredBySpecification('RFC1459', 'RFC2812')
    def testInviteNonExistingChannelTransmitted(self):
        """“There is no requirement that the channel the target user is being
        invited to must exist or be a valid channel.”
        -- <https://tools.ietf.org/html/rfc1459#section-4.2.7>
        and <https://tools.ietf.org/html/rfc2812#section-3.2.7>

        “Only the user inviting and the user being invited will receive
        notification of the invitation.”
        -- <https://tools.ietf.org/html/rfc2812#section-3.2.7>
        """
        self.connectClient('foo')
        self.connectClient('bar')
        self.getMessages(1)
        self.getMessages(2)
        self.sendLine(1, 'INVITE #chan bar')
        self.getMessages(1)
        l = self.getMessages(2)
        self.assertNotEqual(l, [],
                fail_msg='After using “INVITE #chan bar” while #chan does '
                'not exist, “bar” received nothing.')
        self.assertMessageEqual(l[0], command='INVITE',
                params=['#chan', 'bar'],
                fail_msg='After “foo” invited “bar” do non-existing channel '
                '#chan, “bar” should have received “INVITE #chan bar” but '
                'got this instead: {msg}')

    @cases.SpecificationSelector.requiredBySpecification('RFC1459', 'RFC2812')
    def testInviteNonExistingChannelEchoed(self):
        """“There is no requirement that the channel the target user is being
        invited to must exist or be a valid channel.”
        -- <https://tools.ietf.org/html/rfc1459#section-4.2.7>
        and <https://tools.ietf.org/html/rfc2812#section-3.2.7>

        “Only the user inviting and the user being invited will receive
        notification of the invitation.”
        -- <https://tools.ietf.org/html/rfc2812#section-3.2.7>
        """
        self.connectClient('foo')
        self.connectClient('bar')
        self.getMessages(1)
        self.getMessages(2)
        self.sendLine(1, 'INVITE #chan bar')
        l = self.getMessages(1)
        self.assertNotEqual(l, [],
                fail_msg='After using “INVITE #chan bar” while #chan does '
                'not exist, the author received nothing.')
        self.assertMessageEqual(l[0], command='INVITE',
                params=['#chan', 'bar'],
                fail_msg='After “foo” invited “bar” do non-existing channel '
                '#chan, “foo” should have received “INVITE #chan bar” but '
                'got this instead: {msg}')

class testChannelCaseSensitivity(cases.BaseServerTestCase):
    def _testChannelsEquivalent(name1, name2):
        @cases.SpecificationSelector.requiredBySpecification('RFC1459', 'RFC2812',
                strict=True)
        def f(self):
            self.connectClient('foo')
            self.connectClient('bar')
            self.joinClient(1, name1)
            self.joinClient(2, name2)
            try:
                m = self.getMessage(1)
                self.assertMessageEqual(m, command='JOIN',
                        nick='bar')
            except client_mock.NoMessageException:
                raise AssertionError(
                        'Channel names {} and {} are not equivalent.'
                        .format(name1, name2))
        f.__name__ = 'testEquivalence__{}__{}'.format(name1, name2)
        return f
    def _testChannelsNotEquivalent(name1, name2):
        @cases.SpecificationSelector.requiredBySpecification('RFC1459', 'RFC2812',
                strict=True)
        def f(self):
            self.connectClient('foo')
            self.connectClient('bar')
            self.joinClient(1, name1)
            self.joinClient(2, name2)
            try:
                m = self.getMessage(1)
            except client_mock.NoMessageException:
                pass
            else:
                self.assertMessageEqual(m, command='JOIN',
                        nick='bar') # This should always be true
                raise AssertionError(
                        'Channel names {} and {} are equivalent.'
                        .format(name1, name2))
        f.__name__ = 'testEquivalence__{}__{}'.format(name1, name2)
        return f

    testSimpleEquivalent = _testChannelsEquivalent('#Foo', '#foo')
    testSimpleNotEquivalent = _testChannelsNotEquivalent('#Foo', '#fooa')
    testFancyEquivalent = _testChannelsEquivalent('#F]|oo{', '#f}\\oo[')
    testFancyNotEquivalent = _testChannelsEquivalent('#F}o\\o[', '#f]o|o{')



import ecdsa
import base64
import pyxmpp2_scram as scram

from irctest import cases
from irctest import authentication
from irctest.irc_utils.message_parser import Message

ECDSA_KEY = """
-----BEGIN EC PARAMETERS-----
BggqhkjOPQMBBw==
-----END EC PARAMETERS-----
-----BEGIN EC PRIVATE KEY-----
MHcCAQEEIIJueQ3W2IrGbe9wKdOI75yGS7PYZSj6W4tg854hlsvmoAoGCCqGSM49
AwEHoUQDQgAEAZmaVhNSMmV5r8FXPvKuMnqDKyIA9pDHN5TNMfiF3mMeikGgK10W
IRX9cyi2wdYg9mUUYyh9GKdBCYHGUJAiCA==
-----END EC PRIVATE KEY-----
"""

class SaslTestCase(cases.BaseClientTestCase, cases.ClientNegociationHelper,
                   cases.OptionalityHelper):
    @cases.OptionalityHelper.skipUnlessHasMechanism('PLAIN')
    def testPlain(self):
        """Test PLAIN authentication with correct username/password."""
        auth = authentication.Authentication(
                mechanisms=[authentication.Mechanisms.plain],
                username='jilles',
                password='sesame',
                )
        m = self.negotiateCapabilities(['sasl'], auth=auth)
        self.assertEqual(m, Message([], None, 'AUTHENTICATE', ['PLAIN']))
        self.sendLine('AUTHENTICATE +')
        m = self.getMessage()
        self.assertEqual(m, Message([], None, 'AUTHENTICATE',
            ['amlsbGVzAGppbGxlcwBzZXNhbWU=']))
        self.sendLine('900 * * jilles :You are now logged in.')
        self.sendLine('903 * :SASL authentication successful')
        m = self.negotiateCapabilities(['sasl'], False)
        self.assertEqual(m, Message([], None, 'CAP', ['END']))

    @cases.OptionalityHelper.skipUnlessHasMechanism('PLAIN')
    def testPlainNotAvailable(self):
        """`sasl=EXTERNAL` is advertized, whereas the client is configured
        to use PLAIN.

        A client implementing sasl-3.2 can give up authentication immediately.
        A client not implementing it will try authenticating, and will get
        a 904.
        """
        auth = authentication.Authentication(
                mechanisms=[authentication.Mechanisms.plain],
                username='jilles',
                password='sesame',
                )
        m = self.negotiateCapabilities(['sasl=EXTERNAL'], auth=auth)
        self.assertEqual(self.acked_capabilities, {'sasl'})
        if m == Message([], None, 'CAP', ['END']):
            # IRCv3.2-style
            return
        self.assertEqual(m, Message([], None, 'AUTHENTICATE', ['PLAIN']))
        self.sendLine('904 {} :SASL auth failed'.format(self.nick))
        m = self.getMessage()
        self.assertMessageEqual(m, command='CAP')


    @cases.OptionalityHelper.skipUnlessHasMechanism('PLAIN')
    def testPlainLarge(self):
        """Test the client splits large AUTHENTICATE messages whose payload
        is not a multiple of 400.
        <http://ircv3.net/specs/extensions/sasl-3.1.html#the-authenticate-command>
        """
        # TODO: authzid is optional
        auth = authentication.Authentication(
                mechanisms=[authentication.Mechanisms.plain],
                username='foo',
                password='bar'*200,
                )
        authstring = base64.b64encode(b'\x00'.join(
            [b'foo', b'foo', b'bar'*200])).decode()
        m = self.negotiateCapabilities(['sasl'], auth=auth)
        self.assertEqual(m, Message([], None, 'AUTHENTICATE', ['PLAIN']))
        self.sendLine('AUTHENTICATE +')
        m = self.getMessage()
        self.assertEqual(m, Message([], None, 'AUTHENTICATE',
            [authstring[0:400]]), m)
        m = self.getMessage()
        self.assertEqual(m, Message([], None, 'AUTHENTICATE',
            [authstring[400:800]]))
        m = self.getMessage()
        self.assertEqual(m, Message([], None, 'AUTHENTICATE',
            [authstring[800:]]))
        self.sendLine('900 * * {} :You are now logged in.'.format('foo'))
        self.sendLine('903 * :SASL authentication successful')
        m = self.negotiateCapabilities(['sasl'], False)
        self.assertEqual(m, Message([], None, 'CAP', ['END']))

    @cases.OptionalityHelper.skipUnlessHasMechanism('PLAIN')
    def testPlainLargeMultiple(self):
        """Test the client splits large AUTHENTICATE messages whose payload
        is a multiple of 400.
        <http://ircv3.net/specs/extensions/sasl-3.1.html#the-authenticate-command>
        """
        # TODO: authzid is optional
        auth = authentication.Authentication(
                mechanisms=[authentication.Mechanisms.plain],
                username='foo',
                password='quux'*148,
                )
        authstring = base64.b64encode(b'\x00'.join(
            [b'foo', b'foo', b'quux'*148])).decode()
        m = self.negotiateCapabilities(['sasl'], auth=auth)
        self.assertEqual(m, Message([], None, 'AUTHENTICATE', ['PLAIN']))
        self.sendLine('AUTHENTICATE +')
        m = self.getMessage()
        self.assertEqual(m, Message([], None, 'AUTHENTICATE',
            [authstring[0:400]]), m)
        m = self.getMessage()
        self.assertEqual(m, Message([], None, 'AUTHENTICATE',
            [authstring[400:800]]))
        m = self.getMessage()
        self.assertEqual(m, Message([], None, 'AUTHENTICATE',
            ['+']))
        self.sendLine('900 * * {} :You are now logged in.'.format('foo'))
        self.sendLine('903 * :SASL authentication successful')
        m = self.negotiateCapabilities(['sasl'], False)
        self.assertEqual(m, Message([], None, 'CAP', ['END']))

    @cases.OptionalityHelper.skipUnlessHasMechanism('ECDSA-NIST256P-CHALLENGE')
    def testEcdsa(self):
        """Test ECDSA authentication.
        """
        auth = authentication.Authentication(
                mechanisms=[authentication.Mechanisms.ecdsa_nist256p_challenge],
                username='jilles',
                ecdsa_key=ECDSA_KEY,
                )
        m = self.negotiateCapabilities(['sasl'], auth=auth)
        self.assertEqual(m, Message([], None, 'AUTHENTICATE', ['ECDSA-NIST256P-CHALLENGE']))
        self.sendLine('AUTHENTICATE +')
        m = self.getMessage()
        self.assertEqual(m, Message([], None, 'AUTHENTICATE',
            ['amlsbGVz'])) # jilles
        self.sendLine('AUTHENTICATE Zm9vYmFy') # foobar
        m = self.getMessage()
        self.assertMessageEqual(m, command='AUTHENTICATE')
        sk = ecdsa.SigningKey.from_pem(ECDSA_KEY)
        vk = sk.get_verifying_key()
        signature = base64.b64decode(m.params[0])
        try:
            vk.verify(signature, b'foobar')
        except ecdsa.BadSignatureError:
            raise AssertionError('Bad signature')
        self.sendLine('900 * * foo :You are now logged in.')
        self.sendLine('903 * :SASL authentication successful')
        m = self.negotiateCapabilities(['sasl'], False)
        self.assertEqual(m, Message([], None, 'CAP', ['END']))

    @cases.OptionalityHelper.skipUnlessHasMechanism('SCRAM-SHA-256')
    def testScram(self):
        """Test SCRAM-SHA-256 authentication.
        """
        auth = authentication.Authentication(
                mechanisms=[authentication.Mechanisms.scram_sha_256],
                username='jilles',
                password='sesame',
                )
        class PasswdDb:
            def get_password(self, *args):
                return ('sesame', 'plain')
        authenticator = scram.SCRAMServerAuthenticator('SHA-256',
                channel_binding=False, password_database=PasswdDb())

        m = self.negotiateCapabilities(['sasl'], auth=auth)
        self.assertEqual(m, Message([], None, 'AUTHENTICATE', ['SCRAM-SHA-256']))
        self.sendLine('AUTHENTICATE +')

        m = self.getMessage()
        self.assertEqual(m.command, 'AUTHENTICATE', m)
        client_first = base64.b64decode(m.params[0])
        response = authenticator.start(properties={}, initial_response=client_first)
        assert isinstance(response, bytes), response
        self.sendLine('AUTHENTICATE :' + base64.b64encode(response).decode())

        m = self.getMessage()
        self.assertEqual(m.command, 'AUTHENTICATE', m)
        msg = base64.b64decode(m.params[0])
        r = authenticator.response(msg)
        assert isinstance(r, tuple), r
        assert len(r) == 2, r
        (properties, response) = r
        self.sendLine('AUTHENTICATE :' + base64.b64encode(response).decode())
        self.assertEqual(properties, {'authzid': None, 'username': 'jilles'})

    @cases.OptionalityHelper.skipUnlessHasMechanism('SCRAM-SHA-256')
    def testScramBadPassword(self):
        """Test SCRAM-SHA-256 authentication with a bad password.
        """
        auth = authentication.Authentication(
                mechanisms=[authentication.Mechanisms.scram_sha_256],
                username='jilles',
                password='sesame',
                )
        class PasswdDb:
            def get_password(self, *args):
                return ('notsesame', 'plain')
        authenticator = scram.SCRAMServerAuthenticator('SHA-256',
                channel_binding=False, password_database=PasswdDb())

        m = self.negotiateCapabilities(['sasl'], auth=auth)
        self.assertEqual(m, Message([], None, 'AUTHENTICATE', ['SCRAM-SHA-256']))
        self.sendLine('AUTHENTICATE +')

        m = self.getMessage()
        self.assertEqual(m.command, 'AUTHENTICATE', m)
        client_first = base64.b64decode(m.params[0])
        response = authenticator.start(properties={}, initial_response=client_first)
        assert isinstance(response, bytes), response
        self.sendLine('AUTHENTICATE :' + base64.b64encode(response).decode())

        m = self.getMessage()
        self.assertEqual(m.command, 'AUTHENTICATE', m)
        msg = base64.b64decode(m.params[0])
        with self.assertRaises(scram.NotAuthorizedException):
            authenticator.response(msg)

class Irc302SaslTestCase(cases.BaseClientTestCase, cases.ClientNegociationHelper,
                         cases.OptionalityHelper):
    @cases.OptionalityHelper.skipUnlessHasMechanism('PLAIN')
    def testPlainNotAvailable(self):
        """Test the client does not try to authenticate using a mechanism the
        server does not advertise.
        Actually, this is optional."""
        auth = authentication.Authentication(
                mechanisms=[authentication.Mechanisms.plain],
                username='jilles',
                password='sesame',
                )
        m = self.negotiateCapabilities(['sasl=EXTERNAL'], auth=auth)
        self.assertEqual(self.acked_capabilities, {'sasl'})
        self.assertEqual(m, Message([], None, 'CAP', ['END']))



import sys
import unittest
import argparse
import unittest
import functools
import importlib
from .cases import _IrcTestCase
from .runner import TextTestRunner
from .specifications import Specifications
from .basecontrollers import BaseClientController, BaseServerController

def main(args):
    try:
        module = importlib.import_module(args.module)
    except ImportError:
        print('Cannot import module {}'.format(args.module), file=sys.stderr)
        exit(1)

    controller_class = module.get_irctest_controller_class()
    if issubclass(controller_class, BaseClientController):
        import irctest.client_tests as module
    elif issubclass(controller_class, BaseServerController):
        import irctest.server_tests as module
    else:
        print(r'{}.Controller should be a subclass of '
                r'irctest.basecontroller.Base{{Client,Server}}Controller'
                .format(args.module),
                file=sys.stderr)
        exit(1)
    _IrcTestCase.controllerClass = controller_class
    _IrcTestCase.show_io = args.show_io
    _IrcTestCase.strictTests = not args.loose
    if args.specification:
        try:
            _IrcTestCase.testedSpecifications = frozenset(
                Specifications.of_name(x) for x in args.specification
                )
        except ValueError:
            print('Invalid set of specifications: {}'
                    .format(', '.join(args.specification)))
            exit(1)
    else:
        _IrcTestCase.testedSpecifications = frozenset(
                Specifications)
    print('Testing {} on specification(s): {}'.format(
        controller_class.software_name,
        ', '.join(sorted(map(lambda x:x.value,
            _IrcTestCase.testedSpecifications)))))
    ts = module.discover()
    testRunner = TextTestRunner(
            verbosity=args.verbose,
            descriptions=True,
            )
    testLoader = unittest.loader.defaultTestLoader
    result = testRunner.run(ts)
    if result.failures or result.errors:
        exit(1)
    else:
        exit(0)


parser = argparse.ArgumentParser(
        description='A script to test interoperability of IRC software.')
parser.add_argument('module', type=str,
        help='The module used to run the tested program.')
parser.add_argument('--show-io', action='store_true',
        help='Show input/outputs with the tested program.')
parser.add_argument('-v', '--verbose', action='count', default=1,
        help='Verbosity. Give this option multiple times to make '
        'it even more verbose.')
parser.add_argument('-s', '--specification', type=str, action='append',
        help=('The set of specifications to test the program with. '
        'Valid values: {}. '
        'Use this option multiple times to test with multiple '
        'specifications. If it is not given, defaults to all.')
        .format(', '.join(x.value for x in Specifications)))
parser.add_argument('-l', '--loose', action='store_true',
        help='Disables strict checks of conformity to the specification. '
        'Strict means the specification is unclear, and the most restrictive '
        'interpretation is choosen.')


args = parser.parse_args()
main(args)
